# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Human_Powered/compute_operating_empty_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE
from RCAIDE.Framework.Core import Data 
from .  import compute_fuselage_weight
from .  import compute_tail_weight
from .  import compute_wing_weight 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Vertical Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_empty_weight(vehicle,settings=None):
    """ Computes weights estimates for human powered aircraft
    
    Assumptions:
       All of this is from AIAA 89-2048, units are in kg. These weight estimates
       are from the MIT Daedalus and are valid for very lightweight
       carbon fiber composite structures. This may need to be solved iteratively since
       gross weight is an input.
       
    Source: 
        MIT Daedalus
                       
    Inputs:
        wing - a data dictionary with the fields:
            Sw -       wing area                                                       [meters**2]
            bw -       wing span                                                       [meters]
            cw -       average wing chord                                              [meters]
            deltaw -   average rib spacing to average chord ratio                      [dimensionless]
            Nwr -      number of wing surface ribs (bw**2)/(deltaw*Sw)                 [dimensionless]
            t_cw -     wing airfoil thickness to chord ratio                           [dimensionless]
            Nwer -     number of wing end ribs (2*number of individual wing panels -2) [dimensionless]
            
        horizontal - a data dictionary with the fields:
            Sts -      tail surface area                                               [meters]
            bts -      tail surface span                                               [meters]
            cts -      average tail surface chord                                      [meters]
            deltawts - average rib spacing to average chord ratio                      [dimensionless]
            Ntsr -     number of tail surface ribs (bts^2)/(deltats*Sts)               [dimensionless]
            t_cts -    tail airfoil thickness to chord ratio                           [dimensionless]
            
        vertical - a data dictionary with the fields:
            Sts -      tail surface area                                               [meters]
            bts -      tail surface span                                               [meters]
            cts -      average tail surface chord                                      [meters]
            deltawts - average rib spacing to average chord ratio                      [dimensionless]
            Ntsr -     number of tail surface ribs (bts**2)/(deltats*Sts)              [dimensionless]
            t_cts -    tail airfoil thickness to chord ratio                           [dimensionless]
            
        aircraft - a data dictionary with the fields:    
            nult -     ultimate load factor                                            [dimensionless]
            GW -       aircraft gross weight                                           [kilogram]
            qm -       dynamic pressure at maneuvering speed                           [Pascals]
            Ltb -      tailboom length                                                 [meters]
    
    Outputs:
        Wws -      weight of wing spar                                                 [kilogram]
        Wtss -     weight of tail surface spar                                         [kilogram]
        Wwr -      weight of wing ribs                                                 [kilogram]
        Wtsr -     weight of tail surface ribs                                         [kilogram]
        Wwer -     weight of wing end ribs                                             [kilogram]
        WwLE -     weight of wing leading edge                                         [kilogram]
        WtsLE -    weight of tail surface leading edge                                 [kilogram]
        WwTE -     weight of wing trailing edge                                        [kilogram]
        Wwc -      weight of wing covering                                             [kilogram]
        Wtsc -     weight of tail surface covering                                     [kilogram]
        Wtb -      tailboom weight                                                     [kilogram]
                
    Properties Used:
        N/A
    """ 
    
    #Unpack
    
    nult   = vehicle.flight_envelope.ultimate_load
    gw     = vehicle.mass_properties.max_takeoff
    qm     = vehicle.flight_envelope.maximum_dynamic_pressure
    
    for wing in vehicle.wings:
        if isinstance(wing,RCAIDE.Library.Components.Wings.Main_Wing):
            Sw      = wing.areas.reference
            bw      = wing.spans.projected
            cw      = wing.chords.mean_aerodynamic
            Nwr     = wing.number_ribs
            t_cw    = wing.thickness_to_chord
            Nwer    = wing.number_end_ribs
            W_wing = compute_wing_weight(Sw,bw,cw,Nwr,t_cw,Nwer,nult,gw)
            wing.mass_properties.mass = W_wing
    
        # Horizontal Tail weight
        elif isinstance(wing,RCAIDE.Library.Components.Wings.Horizontal_Tail): 
            S_h    = wing.areas.reference
            b_h    = wing.spans.projected
            chs    = wing.chords.mean_aerodynamic
            Nhsr   = wing.number_ribs
            t_ch   = wing.thickness_to_chord
            W_ht  = compute_tail_weight(S_h,b_h,chs,Nhsr,t_ch,qm)
            wing.mass_properties.mass = W_ht

        # Vertical Tail weight 
        elif isinstance(wing,RCAIDE.Library.Components.Wings.Vertical_Tail):     
            S_v    = wing.areas.reference
            b_v    = wing.spans.projected
            cvs    = wing.chords.mean_aerodynamic
            Nvsr   = wing.number_ribs
            t_cv   = wing.thickness_to_chord
            W_vt   = compute_tail_weight(S_v,b_v,cvs,Nvsr,t_cv,qm)
            wing.mass_properties.mass = W_vt

    for fuselage in vehicle.fuselages: 
        Ltb     = vehicle.Ltb  
        W_tb   = compute_fuselage_weight(S_h,qm,Ltb)
        fuselage.mass_properties.mass = W_tb
    
    output                                  = Data()
    output.empty                            = Data()  
    output.empty.structural                 =  Data()
    output.empty.structural.wings           = W_wing +  W_ht +  W_vt
    output.empty.structural.fuselage        = W_tb  
    output.empty.total = W_ht + W_tb + W_vt + W_wing
    
    return output


# RCAIDE/Library/Attributes/Propellants/Methane.py
#  
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant 

# ---------------------------------------------------------------------------------------------------------------------- 
# Methane Propellant Class
# ----------------------------------------------------------------------------------------------------------------------  
class Methane(Propellant):
    """Methane class propellant  
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Density at -162C, 1 atm
        
        Source:  
        """    
        self.tag                       = 'Methane'
        self.reactant                  = 'O2'
        self.density                   = 422.6                            # kg/m^3 (15 C, 1 atm)
        self.specific_energy           = 5.34e7                           # J/kg
        self.energy_density            = 2.26e10                          # J/m^3
        self.lower_heating_value       = 5.0e7                            # J/kg  
        

        self.use_high_fidelity_kinetics_model      =  False 
        self.fuel_surrogate_chemical_properties    = {'CH4':1.0}
        self.fuel_chemical_properties              = {'CH4':1.0}      # [2] More accurate kinetic mechanism, slower simulation    
        self.air_chemical_properties               = {'O2':0.2095, 'N2':0.7809, 'AR':0.0096}
        self.surrogate_species_list                = ['CO', 'CO2', 'H2O']
        self.species_list                          = ['CO', 'CO2', 'H2O', 'NO', 'NO2', 'CSOLID']   
        self.surrogate_chemical_kinetics           = 'Fuel_Surrogate.yaml'
        self.chemical_kinetics                     = 'Fuel.yaml'
        self.oxidizer                              = 'Air.yaml'        
        
        self.global_warming_potential_100.CO2       = 1     # CO2e/kg  
        self.global_warming_potential_100.H2O       = 0.06  # CO2e/kg  
        self.global_warming_potential_100.SO2       = -226  # CO2e/kg  
        self.global_warming_potential_100.NOx       = 52    # CO2e/kg  
        self.global_warming_potential_100.Soot      = 1166  # CO2e/kg    
        self.global_warming_potential_100.Contrails = 11    # kg/CO2e/km          


# RCAIDE/Library/Components/Thermal_Management/Heat_Exchangers/Cross_flow_Heat_Exchanger.py
# 
# Created:  Apr 2024, S. Shekar 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports
from RCAIDE.Framework.Core                                                                import Data, Units 
from RCAIDE.Library.Components                                                            import Component  
from RCAIDE.Library.Attributes.Coolants.Glycol_Water                                      import Glycol_Water  
from RCAIDE.Library.Attributes.Gases                                                      import Air
from RCAIDE.Library.Methods.Thermal_Management.Heat_Exchangers.Cross_Flow_Heat_Exchanger  import  cross_flow_hex_rating_model, append_cross_flow_heat_exchanger_conditions, append_cross_flow_hex_segment_conditions
from RCAIDE.Library.Plots.Thermal_Management.plot_cross_flow_heat_exchanger_conditions    import plot_cross_flow_heat_exchanger_conditions 

import os
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Cross Flow Heat Exchanger 
# ----------------------------------------------------------------------------------------------------------------------  
class Cross_Flow_Heat_Exchanger(Component):
    """
    A class representing a cross-flow heat exchanger with strip fins for thermal 
    management systems.

    Attributes
    ----------
    tag : str
        Unique identifier for the heat exchanger, defaults to 'cross_flow_heat_exchanger'
        
    coolant : Glycol_Water
        Coolant properties for calculations, defaults to Glycol_Water()
        
    air : Air
        Air properties for calculations, defaults to Air()
        
    design_heat_removed : float
        Design heat removal capacity, defaults to 0.0
        
    minimum_air_speed : float
        Minimum air velocity through exchanger, defaults to 105 knots
        
    heat_exchanger_efficiency : float
        Overall heat transfer effectiveness, defaults to 0.8381
        
    density : float
        Material density in kg/m^3, defaults to 8440
        
    thermal_conductivity : float
        Material thermal conductivity in W/m.K, defaults to 121
        
    specific_heat_capacity : float
        Material specific heat in J/kg.K, defaults to 871
        
    stack_length : float
        Length of heat exchanger core, defaults to 1
        
    stack_width : float
        Width of heat exchanger core, defaults to 1
        
    stack_height : float
        Height of heat exchanger core, defaults to 1
        
    t_w : float
        Plate thickness in meters, defaults to 5e-4
        
    t_f : float
        Fin thickness in meters, defaults to 1e-4
        
    fin_spacing_cold : float
        Cold side fin spacing in meters, defaults to 2.54e-3
        
    fin_spacing_hot : float
        Hot side fin spacing in meters, defaults to 2.54e-3
        
    fin_metal_thickness_hot : float
        Hot side fin metal thickness in meters, defaults to 0.102e-3
        
    fin_metal_thickness_cold : float
        Cold side fin metal thickness in meters, defaults to 0.102e-3
        
    fin_exposed_strip_edge_hot : float
        Hot side exposed strip edge in meters, defaults to 3.175e-3
        
    fin_exposed_strip_edge_cold : float
        Cold side exposed strip edge in meters, defaults to 3.175e-3
        
    fin_area_density_hot : float
        Hot side fin area density in m^2/m^3, defaults to 2254
        
    fin_area_density_cold : float
        Cold side fin area density in m^2/m^3, defaults to 2254
        
    finned_area_to_total_area_hot : float
        Hot side fin area ratio, defaults to 0.785
        
    finned_area_to_total_area_cold : float
        Cold side fin area ratio, defaults to 0.785
        
    coolant_hydraulic_diameter : float
        Coolant passage hydraulic diameter in meters, defaults to 1.54e-3
        
    air_hydraulic_diameter : float
        Air passage hydraulic diameter in meters, defaults to 1.54e-3
        
    fin_conductivity : float
        Fin thermal conductivity in W/m.K, defaults to 121
        
    wall_conductivity : float
        Wall thermal conductivity in W/m.K, defaults to 121

    Notes
    -----
    The cross-flow heat exchanger uses strip fins and is based on the 1/8-19.86 
    surface designation. The design includes:
    
    * Liquid coolant passages with strip fins
    * Air passages with strip fins
    * Counter-flow arrangement for maximum effectiveness

    **Assumptions**
    
    * Coolant is 50-50 Glycol Water mixture
    * Constant entrance and exit loss coefficients
    * Surface designation of 1/8-19.86 with strip fins

    **Definitions**

    'Strip Fin'
        Interrupted fin design that enhances heat transfer while reducing pressure drop
        
    'Hydraulic Diameter'
        Characteristic dimension for internal flow calculations

    References
    ----------
    [1] Kays, W.M. and London, A.L. (1998) Compact Heat Exchangers. 3rd Edition, 
        McGraw-Hill, New York.

    See Also
    --------
    RCAIDE.Library.Components.Thermal_Management.Heat_Exchangers.Cryogenic_Heat_Exchanger
        Alternative heat exchanger for cryogenic applications
    """

    def __defaults__(self):
        """
        Sets default values for the cross-flow heat exchanger attributes.
        """         
        self.tag                                                    = 'cross_flow_heat_exchanger'
        self.coolant                                                = Glycol_Water() 
        self.air                                                    = Air() 
        self.design_heat_removed                                    = 0.0
        self.minimum_air_speed                                      = 105 * Units.knots 
        
        # heat exchanger: thermophysical properties                        
        self.heat_exchanger_efficiency                              = 0.8381
        self.density                                                = 8440   # kg/m^3
        self.thermal_conductivity                                   = 121    # W/m.K
        self.specific_heat_capacity                                 = 871    # J/kg.K 


        self.stack_length                                           = 1
        self.stack_width                                            = 1
        self.stack_height                                           = 1
                         
        self.design_air_flow_rate                                   = None 
        self.design_air_inlet_pressure                              = None 
        self.design_coolant_inlet_pressure                          = None 

        # heat exchanger: geometric properties      
        # Plate thickness                       
        self.t_w                                                    = 5e-4   # m

        # Fin Thickness
        self.t_f                                                    = 1e-4   # m

        self.fin_spacing_cold                                       = 2.54e-3 #m
        self.fin_spacing_hot                                        = 2.54e-3 #m

        # Fin metal thickness
        self.fin_metal_thickness_hot                                = 0.102e-3 #m
        self.fin_metal_thickness_cold                               = 0.102e-3 #m

        # Strip edge exposed 
        self.fin_exposed_strip_edge_hot                             = 3.175e-3 #m
        self.fin_exposed_strip_edge_cold                            = 3.175e-3 #m

        # Finned area density 
        self.fin_area_density_hot                                   = 2254 # m^2/m^3 
        self.fin_area_density_cold                                  = 2254 # m^2/m^3 

        # Ratio of finned area to total area 
        self.finned_area_to_total_area_hot                          = 0.785
        self.finned_area_to_total_area_cold                         = 0.785

        # Hydraullic Diameter
        self.coolant_hydraulic_diameter                             = 1.54e-3  #m
        self.air_hydraulic_diameter                                 = 1.54e-3  #m

        # Fin and wall Conductivity 
        self.fin_conductivity                                        = 121    # W/m.K
        self.wall_conductivity                                       = 121    # W/m.K        

        # Fan
        self.fan                                                    = Data() # Replace with RCAIDE Fan 
        self.fan.efficiency                                         = 0.7
        self.fan.active                                             = True 
        
        # Pump 
        self.pump                                                   = Data() # Replace with RCAIDE Fan 
        self.pump.efficiency                                        = 0.7 

        # Operating Conditions 
        self.percent_operation                                      = 1.0
        self.atmospheric_air_inlet_to_outlet_area_ratio             = 0.5 
        self.duct_losses                                            = 0.98 # Inlet, Duct and Nozzle Losses
        self.fan_operation                                          = True


        # Limiting Pressure Drop
        self.pressure_drop_hot                                      = 9.05e3 #Pa 
        self.pressure_drop_cold                                     = 8.79e3 #Pa 


        # Enterance and Exit pressure loss coefficients  
        self.kc_values                                              = load_kc_values()
        self.ke_values                                              = load_ke_values()
        return  

    def append_operating_conditions(self, segment, coolant_line):
        """
        Adds operating conditions for the heat exchanger to a mission segment.

        Parameters
        ----------
        segment : Data
            Mission segment to which conditions are being added
        coolant_line : Data
            Cooling system flow path information
        """
        append_cross_flow_heat_exchanger_conditions(self, segment, coolant_line)
        return
  
    def append_segment_conditions(self, segment, bus, coolant_line, conditions):
        """
        Adds specific segment conditions to the heat exchanger analysis.

        Parameters
        ----------
        segment : Data
            Mission segment being analyzed
        bus : Data
            Electrical bus data
        coolant_line : Data
            Cooling system flow path information
        conditions : Data
            Operating conditions for the segment
        """
        append_cross_flow_hex_segment_conditions(self, segment, bus, coolant_line, conditions)
        return
       
    def compute_heat_exchanger_performance(self, state, bus, coolant_line, delta_t, t_idx):
        """
        Calculates thermal performance of the heat exchanger.

        Parameters
        ----------
        state : Data
            Current system state
        bus : Data
            Electrical bus data
        coolant_line : Data
            Cooling system flow path information
        delta_t : float
            Time step size
        t_idx : int
            Time index in the simulation
        """
        cross_flow_hex_rating_model(self, state, bus, coolant_line, delta_t, t_idx)
        return

    def plot_operating_conditions(self, results, coolant_line, save_filename, save_figure, 
                                show_legend, file_type, width, height):
        """
        Creates visualization plots of the heat exchanger operating conditions.

        Parameters
        ----------
        results : Data
            Simulation results data
        coolant_line : Data
            Cooling system flow path information
        save_filename : str
            Path for saving the plot
        save_figure : bool
            Flag to save the figure
        show_legend : bool
            Flag to display plot legend
        file_type : str
            Output file format
        width : float
            Plot width
        height : float
            Plot height
        """
        plot_cross_flow_heat_exchanger_conditions(self, results, coolant_line, save_filename,
                                                save_figure, show_legend, file_type, width, height)     
        return    

    def load_kc_values(): 
        """
        Loads entrance loss coefficient data from file.

        Returns
        -------
        ndarray
            Array of entrance loss coefficients
        """
        ospath    = os.path.abspath(__file__)
        separator = os.path.sep
        rel_path  = os.path.dirname(ospath) + separator   
        x         = np.loadtxt(rel_path + 'rectangular_passage_Kc.csv', dtype=float, 
                             delimiter=',', comments='Kc') 
        return x 
    
    def load_ke_values():  
        """
        Loads exit loss coefficient data from file.

        Returns
        -------
        ndarray
            Array of exit loss coefficients
        """
        ospath    = os.path.abspath(__file__)
        separator = os.path.sep
        rel_path  = os.path.dirname(ospath) + separator 
        x         = np.loadtxt(rel_path +'rectangular_passage_Ke.csv', dtype=float, 
                             delimiter=',', comments='Ke')
        return x 


# RCAIDE/Library/Attributes/Cryogens/Liquid_Hydrogen.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Cryogen import Cryogen

# ---------------------------------------------------------------------------------------------------------------------- 
#  Liquid Hydrogen Cryogen
# ----------------------------------------------------------------------------------------------------------------------  
class Liquid_Hydrogen(Cryogen):
    """Liquid Hydrogen Cryogen Class 
    """

    def __defaults__(self):
        """This sets the default values.
        Assumptions:
            None
        
        Source:
            Ekin - Experimental Techniques for Low Temperature Measurements, ISBN 0-19-857054-6
            NIST Chemistry Webbook
       """ 
        
        self.tag                        = 'Liquid_Hydrogen'
        self.density                    =    59.9            # [kg/m^3] 
        self.specific_energy            =   141.86e6         # [J/kg] 
        self.energy_density             =  8491.0e6          # [J/m^3]
        self.temperatures.freeze        =    13.99           # [K]
        self.temperatures.boiling       =    20.271          # [K]
        self.vaporization_enthalpy      =   461.             # [kJ/kg]
        self.specific_heat              =    10.67           # [kJ/kgK]

        # Coefficiencts for polynomial fit of Liquid Specific Heat Capacity (C_P) curve.
        # C_P = CP_C3*T^3 + CP_C2*T^2 + CP_C1*T^1 + CP_C0*T^0 where C_P is Specific Heat Capacity (J/gK) T is temperature (kelvin).
        # Data from NIST Chemistry Webbook. Pressure is 1.295MPa.
        self.LCP_C0                     =   -31.2
        self.LCP_C1                     =     5.56
        self.LCP_C2                     =    -0.272
        self.LCP_C3                     =     4.76E-03
        
        # Range for which this polynomial fit holds
        self.LCP_minT                   =    15.0              # [K]
        self.LCP_maxT                   =    30.0              # [K]

        # Coefficiencts for polynomial fit of Gas Specific Heat Capacity (C_P) curve.
        # C_P = CP_C3*T^3 + CP_C2*T^2 + CP_C1*T^1 + CP_C0*T^0 where C_P is Specific Heat Capacity (J/gK) T is temperature (kelvin).
        # Data from NIST Chemistry Webbook. Pressure is 0.01 MPa
        self.GCP_C0                     =   10.3
        self.GCP_C1                     =   -7.39E-03
        self.GCP_C2                     =    0.221E-03
        self.GCP_C3                     =   -0.516E-06
        
        # Range for which this polynomial fit holds
        self.GCP_minT                   =   20.0              # [K]
        self.GCP_maxT                   =  300.0              # [K]

        # Antoine Equation Coefficients for calculatating the evaporation temperature.
        # log10(P) = A - (B/(T+C)) where P is vapour pressure (Pa) and T temperature (kelvin).
        # Data from NIST Chemistry Webbook, coefficients converted so as to use pressure in Pa.
        self.antoine_A                  =    8.54314
        self.antoine_B                  =   99.395
        self.antoine_C                  =    7.726
        
        # Range for which Antoine Equation is referenced
        self.antoine_minT               =   21.01             # [K]
        self.antoine_maxT               =   32.27             # [K]

        # Coefficiencts for polynomial fit of vapourisation enthalpy
        # ΔH = H_C3*P^3 + H_C2*P^2 + H_C1*P^1 + H_C0*P^0 where ΔH is vapourisation enthalpy (kJ/kg), P is pressure (Pa). 
        self.H_C0                       =   464.
        self.H_C1                       =    -0.000176
        self.H_C2                       =    52.3E-12
        self.H_C3                       =  -100.00E-18
        
        # Range for which this polynomial fit holds
        self.H_minP                     =     0.02E6         # [Pa]
        self.H_maxP                     =     1.20E6         # [Pa]




## @ingroup Methods-Performance
# generate_V_n_diagram.py
#
# Created:  Nov 2018, S. Karpuk
# Modified:

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Core import Units 
from RCAIDE.Framework.Mission.Common  import Results  

# package imports
import numpy as np
import matplotlib.pyplot as plt

# ----------------------------------------------------------------------
#  Compute a V-n diagram
# ----------------------------------------------------------------------

## @ingroup Methods-Performance
def generate_V_n_diagram(vehicle,analyses,altitude,delta_ISA):
    
    """ Computes a V-n diagram for a given aircraft and given regulations for ISA conditions

    Source:
    S. Gudmundsson "General Aviation Aircraft Design: Applied Methods and Procedures", Butterworth-Heinemann; 1 edition
    CFR FAR Part 23: https://www.ecfr.gov/cgi-bin/text-idx?SID=0e6a13c7c1de7f501d0eb0a4d71418bd&mc=true&tpl=/ecfrbrowse/Title14/14cfr23_main_02.tpl
    CFR FAR Part 25: https://www.ecfr.gov/cgi-bin/text-idx?tpl=/ecfrbrowse/Title14/14cfr25_main_02.tpl

    Inputs:
    analyses.base.atmosphere               [RCAIDE data type]
    vehicle.
      reference_area                       [m^2]
      maximum_lift_coefficient             [Unitless]
      minimum_lift_coefficient             [Unitless]
      chords.mean_aerodynamic              [m]
      envelope.FARpart_number              [Unitless]
        positive_limit_load               [Unitless]
        negative_limit_load               [Unitless]
        cruise_mach                        [Unitless]
    weight                                 [kg]
    altitude                               [m]
    delta_ISA                              [deg C]

    Outputs:
    V_n_data

    Properties Used:
    N/A

    Description:
    The script creates an aircraft V-n diagram based on the input parameters specified by the user.
    Depending on the certification flag, an appropriate diagram, output and log files are created.
    """
    
    weight =  vehicle.mass_properties.max_takeoff
 
    # ----------------------------------------------
    # Unpack
    # ---------------------------------------------- 
    FAR_part_number = vehicle.flight_envelope.FAR_part_number
    atmo            = analyses.atmosphere
    Mc              = vehicle.flight_envelope.design_mach_number

    for wing in vehicle.wings: 
        reference_area  = vehicle.reference_area 
        Cmac            = wing.chords.mean_aerodynamic 
        pos_limit_load  = vehicle.flight_envelope.positive_limit_load
        neg_limit_load  = vehicle.flight_envelope.positive_limit_load

    category_tag = vehicle.flight_envelope.category
    
    # ----------------------------------------------
    # Computing atmospheric conditions
    # ----------------------------------------------
    atmo_values       = atmo.compute_values(altitude,delta_ISA)
    SL_atmo_values    = atmo.compute_values(0,delta_ISA)
    conditions        = Results()

    rho               = atmo_values.density
    sea_level_rho     = SL_atmo_values.density
    sea_level_gravity = atmo.planet.sea_level_gravity
    Vc                = Mc * atmo_values.speed_of_sound
    
    # ------------------------------
    # Computing lift-curve slope
    # ------------------------------ 
    results =  evalaute_aircraft(vehicle,altitude,Vc) 
    CLa     =  results.segments.cruise.conditions.static_stability.derivatives.Clift_alpha[0, 0] 

    # -----------------------------------------------------------
    # Determining vehicle minimum and maximum lift coefficients
    # -----------------------------------------------------------
    if vehicle.flight_envelope.maximum_lift_coefficient != None:
        maximum_lift_coefficient = vehicle.flight_envelope.maximum_lift_coefficient
    else:
        raise ValueError("Maximum lift coefficient not specified.")

    if vehicle.flight_envelope.minimum_lift_coefficient != None:
        minimum_lift_coefficient = vehicle.flight_envelope.minimum_lift_coefficient
    else: 
        raise ValueError("Minimum lift coefficient not specified.")
             
    # -----------------------------------------------------------------------------
    # Convert all terms to English (Used for FAR) and remove elements from arrays
    # -----------------------------------------------------------------------------
    altitude          = altitude / Units.ft
    rho               = rho[0,0] / Units['slug/ft**3']
    sea_level_rho     = sea_level_rho[0,0] / Units['slug/ft**3']
    density_ratio     = (rho/sea_level_rho)**0.5
    sea_level_gravity = sea_level_gravity / Units['ft/s**2']
    weight            = weight / Units['slug'] * sea_level_gravity
    reference_area    = reference_area / Units['ft**2']
    Cmac              = Cmac / Units.ft
    wing_loading      = weight / reference_area
    Vc                = Vc / Units['ft/s']
    
    load_factors_pos    = np.zeros(shape=(5));
    load_factors_neg    = np.zeros(shape=(5));
    load_factors_pos[1] =  1;
    load_factors_neg[1] = -1;
    airspeeds_pos       = np.zeros(shape=(5));
    airspeeds_neg       = np.zeros(shape=(5))
        
    # --------------------------------------------------
    # Establish limit maneuver load factors n+ and n- 
    # -------------------------------------------------- 
    # CFR Part 25
    # Positive and negative limits
    if FAR_part_number == '25':
        # Positive limit 
        load_factors_pos[2] = 2.1 + 24000 / (weight + 10000)
        if load_factors_pos[2] < 2.5: 
            load_factors_pos[2] = 2.5
        elif load_factors_pos[2] < pos_limit_load:
            load_factors_pos[2] = pos_limit_load
        elif load_factors_pos[2] > 3.8: 
            load_factors_pos[2] = 3.8

        # Negative limit 
        load_factors_neg[2] = neg_limit_load
        if load_factors_neg[2] > -1: 
            load_factors_neg[2] = -1

    elif FAR_part_number == '23':
        if category_tag == 'normal' or category_tag == 'commuter':
            # Positive limit 
            load_factors_pos[2] = 2.1 + 24000 / (weight + 10000)
            if load_factors_pos[2] < 2.5: 
                load_factors_pos[2] = 2.5
            elif load_factors_pos[2] < pos_limit_load:
                load_factors_pos[2] = pos_limit_load
            elif load_factors_pos[2] > 3.8: 
                load_factors_pos[2] = 3.8

            # Negative limit 
            load_factors_neg[2] = -0.4 * load_factors_pos[2]
            
        elif category_tag == 'utility':
            # Positive limit 
            load_factors_pos[2] =  pos_limit_load
            if load_factors_pos[2] < 4.4: 
                load_factors_pos[2] = 4.4
                 
            # Negative limit 
            load_factors_neg[2] = -0.4 * load_factors_pos[2]
            
        elif category_tag == 'acrobatic':
            # Positive limit
            load_factors_pos[2] = pos_limit_load
            if load_factors_pos[2] < 6.0: 
                load_factors_pos[2] = 6.0

            # Negative limit    
            load_factors_neg[2] = -0.5 * load_factors_pos[2]
            
        else:
            raise ValueError("Check the category_tag input. The parameter was not found")
       
    else:
        raise ValueError("Check the FARflag input. The parameter was not found")

    # Check input of the limit load
    if abs(neg_limit_load) > abs(load_factors_neg[2]):
        load_factors_neg[2] = -neg_limit_load

    #----------------------------------------
    # Generate a V-n diagram data structure
    #----------------------------------------
    V_n_data                          = Data()
    V_n_data.limit_loads              = Data()
    V_n_data.limit_loads.dive         = Data()
    V_n_data.load_factors             = Data()
    V_n_data.gust_load_factors        = Data()
    V_n_data.Vb_load_factor           = Data()
    V_n_data.airspeeds                = Data()
    V_n_data.Vs1                      = Data()
    V_n_data.Va                       = Data()
    V_n_data.Vb                       = Data()
    V_n_data.load_factors.positive    = load_factors_pos
    V_n_data.load_factors.negative    = load_factors_neg
    V_n_data.airspeeds.positive       = airspeeds_pos
    V_n_data.airspeeds.negative       = airspeeds_neg
    V_n_data.Vc                       = Vc
    V_n_data.weight                   = weight
    V_n_data.wing_loading             = wing_loading
    V_n_data.altitude                 = altitude
    V_n_data.density                  = rho
    V_n_data.density_ratio            = density_ratio
    V_n_data.reference_area           = reference_area
    V_n_data.maximum_lift_coefficient = maximum_lift_coefficient
    V_n_data.minimum_lift_coefficient = minimum_lift_coefficient
    V_n_data.positive_limit_load      = load_factors_pos[2]
    V_n_data.negative_limit_load      = load_factors_neg[2]
    
    # --------------------------------------------------
    # Computing critical speeds (Va, Vc, Vb, Vd, Vs1)
    # -------------------------------------------------- 
    
    # Calculate Stall and Maneuver speeds 
    stall_maneuver_speeds(V_n_data)  
        
    # convert speeds to KEAS for future calculations 
    convert_keas(V_n_data)

    # unpack modified airspeeds 
    airspeeds_pos = V_n_data.airspeeds.positive
    airspeeds_neg = V_n_data.airspeeds.negative
    Vc            = V_n_data.Vc
    Va_pos        = V_n_data.Va.positive 
    Va_neg        = V_n_data.Va.negative
    Va_pos        = V_n_data.Va.positive 
    Va_neg        = V_n_data.Va.negative 
 
    if Va_neg > Vc and Va_neg > Va_pos: 
        Vc = 1.15 * Va_neg
    elif Va_pos > Vc and Va_neg < Va_pos: 
        Vc = 1.15 * Va_pos
    
    # Gust speeds between Vb and Vc (EAS) and minimum Vc
    miu = 2 * wing_loading / (rho * Cmac * CLa * sea_level_gravity)
    Kg  = 0.88 * miu / (5.3 + miu)
          
    if FAR_part_number == '25':
        if altitude < 15000:
            Uref_cruise = (-0.0008 * altitude + 56)
            Uref_rough  = Uref_cruise
            Uref_dive   = 0.5 * Uref_cruise
        else:
            Uref_cruise = (-0.0005142 * altitude + 51.7133)
            Uref_rough  = Uref_cruise
            Uref_dive   = 0.5 * Uref_cruise
            
        # Minimum Cruise speed Vc_min
        coefs = [1, -Uref_cruise * (2.64 + (Kg * CLa * airspeeds_pos[1]**2)/(498 * wing_loading)), 1.72424 * Uref_cruise**2 - airspeeds_pos[1]**2]
        Vc1   = max(np.roots(coefs))
        
    elif FAR_part_number == '23':           
        if altitude < 20000:
            Uref_cruise = 50;
            Uref_dive   = 25;
        else:
            Uref_cruise = (-0.0008333 * altitude + 66.67);
            Uref_dive = (-0.0004167 * altitude + 33.334);

        if category_tag == 'commuter':
            if altitude < 20000:
                Uref_rough = 66
            else:
                Uref_rough = -0.000933 * altitude + 84.667
        else:
            Uref_rough = Uref_cruise
                       
        # Minimum Cruise speed Vc_min
        if category_tag == 'acrobatic':
            Vc1 = 36 * wing_loading**0.5
            if wing_loading >= 20:
                Vc1 = (-0.0925 * wing_loading + 37.85) * wing_loading **0.5
        else:
            Vc1 = 33 * wing_loading**0.5
            if wing_loading >= 20:       
                Vc1 = (-0.055 * wing_loading + 34.1) * wing_loading **0.5

    # checking input Cruise speed 
    if Vc1 > Vc: 
        Vc = Vc1

    # Dive speed 
    if FAR_part_number == '25':        
        airspeeds_pos[4] = 1.25 * Vc
    elif FAR_part_number == '23':
        if category_tag == 'acrobatic':
            airspeeds_pos[4] = 1.55 * Vc1
            if wing_loading > 20:
                airspeeds_pos[4] = (-0.0025 * wing_loading + 1.6) * Vc1
        elif category_tag == 'utility':
            airspeeds_pos[4] = 1.5 * Vc1
            if wing_loading > 20:
                airspeeds_pos[4] = (-0.001875 * wing_loading + 1.5375) * Vc1           
        else:
            airspeeds_pos[4] = 1.4 * Vc1
            if wing_loading > 20:
                airspeeds_pos[4] = (-0.000625 * wing_loading + 1.4125) * Vc1 
        if airspeeds_pos[4] < 1.15 * Vc: 
            airspeeds_pos[4] = 1.15 * Vc
   
    Vd               = airspeeds_pos[4]
    airspeeds_pos[3] = airspeeds_pos[4]
    airspeeds_neg[3] = Vc
    airspeeds_neg[4] = airspeeds_pos[4]
    
    # complete initial load factors
    load_factors_pos[4] = 0
    load_factors_neg[4] = 0
    if category_tag == 'acrobatic' or category_tag == 'utility':
        load_factors_neg[4] = -1
        
    load_factors_pos[3] = load_factors_pos[2]
    load_factors_neg[3] = load_factors_neg[2]

    # add parameters to the data structure
    V_n_data.load_factors.positive    = load_factors_pos
    V_n_data.load_factors.negative    = load_factors_neg
    V_n_data.airspeeds.positive       = airspeeds_pos
    V_n_data.airspeeds.negative       = airspeeds_neg
    V_n_data.Vd                       = Vd
    V_n_data.Vc                       = Vc

    #------------------------
    # Create Stall lines
    #------------------------ 
    Num_of_points = 20                                            # number of points for the stall line
    upper_bound = 2;
    lower_bound = 1;
    stall_line(V_n_data, upper_bound, lower_bound, Num_of_points, 1)
    stall_line(V_n_data, upper_bound, lower_bound, Num_of_points, 2)

    # ----------------------------------------------
    # Determine Gust loads
    # ---------------------------------------------- 
    V_n_data.gust_data           = Data()
    V_n_data.gust_data.airspeeds = Data()
    
    V_n_data.gust_data.airspeeds.rough_gust  = Uref_rough
    V_n_data.gust_data.airspeeds.cruise_gust = Uref_cruise
    V_n_data.gust_data.airspeeds.dive_gust   = Uref_dive
    
    gust_loads(category_tag, V_n_data, Kg, CLa, Num_of_points, FAR_part_number, 1)
    gust_loads(category_tag, V_n_data, Kg, CLa, Num_of_points, FAR_part_number, 2)

    #----------------------------------------------------------------
    # Finalize the load factors for acrobatic and utility aircraft
    #----------------------------------------------------------------
    if category_tag == 'acrobatic' or category_tag == 'utility':
        V_n_data.airspeeds.negative    = np.append(V_n_data.airspeeds.negative, Vd)
        V_n_data.load_factors.negative = np.append(V_n_data.load_factors.negative, 0)

    # ----------------------------------------------
    # Post-processing the V-n diagram
    # ---------------------------------------------- 
    V_n_data.positive_limit_load = max(V_n_data.load_factors.positive)
    V_n_data.negative_limit_load = min(V_n_data.load_factors.negative) 
    post_processing(category_tag, Uref_rough, Uref_cruise, Uref_dive, V_n_data, vehicle) 
    return V_n_data

      
def evalaute_aircraft(vehicle,altitude,Vc): 
    
    # Set up vehicle configs
    configs  = configs_setup(vehicle)

    # create analyses
    analyses = analyses_setup(configs)

    # mission analyses
    mission  = base_mission_setup(analyses,altitude,Vc) 

    # create mission instances (for multiple types of missions)
    missions = missions_setup(mission) 

    # mission analysis 
    results = missions.base_mission.evaluate() 

    return results
 
def analyses_setup(configs):

    analyses = RCAIDE.Framework.Analyses.Analysis.Container()

    # build a base analysis for each config
    for tag,config in configs.items():
        analysis = base_analysis(config)
        analyses[tag] = analysis

    return analyses

def base_analysis(vehicle):

       # ------------------------------------------------------------------
    #   Initialize the Analyses
    # ------------------------------------------------------------------     
    analyses        = RCAIDE.Framework.Analyses.Vehicle() 

    # ------------------------------------------------------------------
    #  Weights
    # ------------------------------------------------------------------
    weights         = RCAIDE.Framework.Analyses.Weights.Weights_EVTOL()
    weights.vehicle = vehicle
    analyses.append(weights)

    # ------------------------------------------------------------------
    #  Aerodynamics Analysis
    # ------------------------------------------------------------------
    aerodynamics                                      = RCAIDE.Framework.Analyses.Aerodynamics.Vortex_Lattice_Method() 
    aerodynamics.vehicle                              = vehicle  
    aerodynamics.settings.use_surrogate               = False 
    aerodynamics.settings.trim_aircraft               = False 
    analyses.append(aerodynamics)
    
    
    # ------------------------------------------------------------------
    #  Energy
    # ------------------------------------------------------------------
    energy     = RCAIDE.Framework.Analyses.Energy.Energy()
    energy.vehicle = vehicle  
    analyses.append(energy)

    # ------------------------------------------------------------------
    #  Planet Analysis
    # ------------------------------------------------------------------
    planet     = RCAIDE.Framework.Analyses.Planets.Earth()
    analyses.append(planet)

    # ------------------------------------------------------------------
    #  Atmosphere Analysis
    # ------------------------------------------------------------------
    atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere.features.planet = planet.features
    analyses.append(atmosphere)   

    # done!
    return analyses

  
def configs_setup(vehicle): 
    # ------------------------------------------------------------------
    #   Initialize Configurations
    # ------------------------------------------------------------------ 
    configs = RCAIDE.Library.Components.Configs.Config.Container() 
    base_config                                                       = RCAIDE.Library.Components.Configs.Config(vehicle)
    base_config.tag                                                   = 'base'     
    configs.append(base_config)  
    return configs
 
def base_mission_setup(analyses,altitude,Vc):   
    '''
    This sets up the nominal cruise of the aircraft
    '''
     
    mission = RCAIDE.Framework.Mission.Sequential_Segments()
    mission.tag = 'base_mission'
  
    # unpack Segments module
    Segments = RCAIDE.Framework.Mission.Segments

    #   Cruise Segment: constant Speed, constant altitude 
    segment                           = Segments.Untrimmed.Untrimmed()
    segment.analyses.extend( analyses.base )   
    segment.tag                       = "cruise" 
    segment.altitude                  = altitude
    segment.air_speed                 = Vc

    segment.flight_dynamics.force_x   = True    
    segment.flight_dynamics.force_z   = True    
    segment.flight_dynamics.force_y   = True     
    segment.flight_dynamics.moment_y  = True 
    segment.flight_dynamics.moment_x  = True
    segment.flight_dynamics.moment_z  = True
    
    mission.append_segment(segment)     
    
    return mission

def missions_setup(mission): 
 
    missions     = RCAIDE.Framework.Mission.Missions()
    
    # base mission 
    mission.tag  = 'base_mission'
    missions.append(mission)
 
    return missions  


#------------------------------------------------------------------------------------------------------
# USEFUL FUNCTIONS
#------------------------------------------------------------------------------------------------------

def stall_maneuver_speeds(V_n_data):

    """ Computes stall and maneuver speeds for positive and negative halves of the the V-n diagram

    Source:
    S. Gudmundsson "General Aviation Aircraft Design: Applied Methods and Procedures", Butterworth-Heinemann; 1 edition

    Inputs:
    V_n_data.
        airspeeds.positive                      [kts]
            negative                            [kts]
        weight                                  [lb]
        density                                 [slug/ft**3]
        density_ratio                           [Unitless]
        reference_area                          [ft**2]
        maximum_lift_coefficient                [Unitless]
        minimum_lift_coefficient                [Unitless]
        load_factors.positive                   [Unitless]
        load_factors.negative                   [Unitless]

    Outputs:
    V_n_data.
        airspeeds.positive                      [kts]
            negative                            [kts]
        Vs1.positive                            [kts]       
            negative                            [kts]
        Va.positive                             [kts]
            negative                            [kts]
        load_factors.positive                   [Unitless]
            negative                            [Unitless]                 

    Properties Used:
    N/A

    Description:   
    """     

    # Unpack
    airspeeds_pos    = V_n_data.airspeeds.positive
    airspeeds_neg    = V_n_data.airspeeds.negative
    weight           = V_n_data.weight
    rho              = V_n_data.density
    reference_area   = V_n_data.reference_area
    max_lift_coef    = V_n_data.maximum_lift_coefficient
    min_lift_coef    = V_n_data.minimum_lift_coefficient
    load_factors_pos = V_n_data.load_factors.positive
    load_factors_neg = V_n_data.load_factors.negative
    
    # Stall speeds
    airspeeds_pos[1] = (2 * weight / (rho * reference_area * max_lift_coef)) ** 0.5
    airspeeds_neg[1] = (2 * weight / (rho * reference_area * abs(min_lift_coef))) ** 0.5
    airspeeds_pos[0] = airspeeds_pos[1]
    airspeeds_neg[0] = airspeeds_neg[1]

    # Maneuver speeds
    airspeeds_pos[2] = airspeeds_pos[1] * load_factors_pos[2] ** 0.5
    airspeeds_neg[2] = (2 * weight * abs(load_factors_neg[2]) / (rho * reference_area * \
                                                                 abs(min_lift_coef))) ** 0.5
    
    # Pack
    V_n_data.airspeeds.positive           = airspeeds_pos
    V_n_data.airspeeds.negative           = airspeeds_neg
    V_n_data.load_factors.positive        = load_factors_pos
    V_n_data.load_factors.negative        = load_factors_neg
    V_n_data.Vs1.positive                 = airspeeds_pos[1]
    V_n_data.Vs1.negative                 = airspeeds_neg[1]
    V_n_data.Va.positive                  = airspeeds_pos[2]
    V_n_data.Va.negative                  = airspeeds_neg[2]
    
#------------------------------------------------------------------------------------------------------------

def stall_line(V_n_data, upper_bound, lower_bound, Num_of_points, sign_flag):
    
    """ Calculates Stall lines of positive and negative halves of the V-n diagram

    Source:
    S. Gudmundsson "General Aviation Aircraft Design: Applied Methods and Procedures", Butterworth-Heinemann; 1 edition

    Inputs:
    V_n_data.
        airspeeds.positive                  [kts]
            negative                        [kts]
        load_factors.positive               [Unitless]
            negative                        [Unitless]
        weight                              [lb]
        density                             [slug/ft**3]
        density_ratio                       [Unitless]
        lift_coefficient                    [Unitless]
        reference_area                      [ft**2]
    lower_bound                             [Unitless]
    Num_of_points                           [Unitless]
    sign_flag                               [Unitless]

    Outputs:
    V_n_data.
        load_factors.positive               [Unitless]
            negative                        [Unitless]
        airspeeds.positive                  [kts]
            negative                        [kts]

    Properties Used:
    N/A

    Description:   
    """  
    # Unpack
    weight          = V_n_data.weight
    reference_area  = V_n_data.reference_area
    density_ratio   = V_n_data.density_ratio
    rho             = V_n_data.density
    
    if sign_flag == 1:
        load_factors     = V_n_data.load_factors.positive
        airspeeds        = V_n_data.airspeeds.positive
        lift_coefficient = V_n_data.maximum_lift_coefficient

    elif sign_flag == 2:
        load_factors     = V_n_data.load_factors.negative
        airspeeds        = V_n_data.airspeeds.negative
        lift_coefficient = V_n_data.minimum_lift_coefficient
    
    delta = (airspeeds[upper_bound] - airspeeds[lower_bound]) / (Num_of_points + 1)     # Step size
    for i in range(Num_of_points):       
        coef      = lower_bound + i + 1
        airspeeds = np.concatenate((airspeeds[:coef], [airspeeds[lower_bound] + (i + 1) * delta], airspeeds[coef:]))
        Vtas      = airspeeds[coef] / density_ratio * Units.knots / Units['ft/s']      
        if load_factors[1] > 0:
            nl = 0.5 * rho * Vtas**2 * reference_area * lift_coefficient / weight
        else:
            nl = -0.5 * rho * Vtas**2 * reference_area * abs(lift_coefficient) / weight
            
        load_factors = np.concatenate((load_factors[:coef], [nl], load_factors[coef:]))

    # Pack
    if sign_flag == 1:
        V_n_data.load_factors.positive     = load_factors
        V_n_data.airspeeds.positive        = airspeeds

    elif sign_flag == 2:
        V_n_data.load_factors.negative     = load_factors
        V_n_data.airspeeds.negative        = airspeeds
    
    return 
#--------------------------------------------------------------------------------------------------------------

def gust_loads(category_tag, V_n_data, Kg, CLa, Num_of_points, FAR_part_number, sign_flag):

    """ Calculates airspeeds and load factors for gust loads and modifies the V-n diagram

    Source:
    S. Gudmundsson "General Aviation Aircraft Design: Applied Methods and Procedures", Butterworth-Heinemann; 1 edition

    Inputs:
    V_n_data.
        airspeeds.positive                      [kts]
            negative                            [kts]
        load_factors.positive                   [Unitless]
            negative                            [Unitless]
        positive_limit_load                    [Unitless]
            negative                            [Unitless]
        weight                                  [lb]
        wing_loading                            [lb/ft**2]
        reference_area                          [ft**2]
        density                                 [slug/ft**3]
        density_ratio                           [Unitless]
        lift_coefficient                        [Unitless]
        Vc                                      [kts]
        Vd                                      [kts]
    Uref_rough                                  [ft/s]
    Uref_cruise                                 [ft/s]
    Uref_dive                                   [ft/s]
    Num_of_points                               [Unitless]
    sign_flag                                   [Unitless]
  
    
    Outputs:
    V_n_data.
        airspeeds.positive                      [kts]
            negative                            [kts]
        load_factors.positive                   [Unitless]
            .negative                           [Unitless]
        gust_load_factors.positive              [Unitless]
            negative                            [Unitless]

    Properties Used:
    N/A

    Description:
    The function calculates the gust-related load factors at critical speeds (Va, Vc, Vd). Then, if the load factors exceed
    the standart diagram limits, the diagram is modified to include the new limit loads.
    For more details, refer to S. Gudmundsson "General Aviation Aircraft Design: Applied Methods and Procedures"
    """

    # Unpack
    weight          = V_n_data.weight
    wing_loading    = V_n_data.wing_loading
    reference_area  = V_n_data.reference_area
    density         = V_n_data.density
    density_ratio   = V_n_data.density_ratio
    Vc              = V_n_data.Vc
    Vd              = V_n_data.Vd
    
    if sign_flag == 1:
        airspeeds        = V_n_data.airspeeds.positive
        load_factors     = V_n_data.load_factors.positive
        lift_coefficient = V_n_data.maximum_lift_coefficient
        limit_load       = V_n_data.positive_limit_load
        Uref_rough       = V_n_data.gust_data.airspeeds.rough_gust
        Uref_cruise      = V_n_data.gust_data.airspeeds.cruise_gust
        Uref_dive        = V_n_data.gust_data.airspeeds.dive_gust

    elif sign_flag == 2:
        airspeeds        = V_n_data.airspeeds.negative
        load_factors     = V_n_data.load_factors.negative
        lift_coefficient = V_n_data.minimum_lift_coefficient
        limit_load       = V_n_data.negative_limit_load
        Uref_rough       = -V_n_data.gust_data.airspeeds.rough_gust
        Uref_cruise      = -V_n_data.gust_data.airspeeds.cruise_gust
        Uref_dive        = -V_n_data.gust_data.airspeeds.dive_gust
        

    gust_load_factors    = np.zeros(shape=(4));    
    gust_load_factors[0] = 1;
    
    # Cruise speed Gust loads at Va and Vc
    gust_load_factors[1] = 1 + Kg * CLa * airspeeds[Num_of_points+2] * Uref_rough / (498 * wing_loading)
    gust_load_factors[2] = 1 + Kg * CLa * Vc * Uref_cruise/(498 * wing_loading)

    # Intersection between cruise gust load and Va
    if abs(gust_load_factors[1]) > abs(limit_load):
        sea_level_rho      = density / density_ratio**2
        coefs              = [709.486 * sea_level_rho * lift_coefficient, -Kg * Uref_rough * CLa, -498 * wing_loading]
        V_inters           = max(np.roots(coefs))
        load_factor_inters = 1 + Kg * CLa * V_inters * Uref_rough / (498 * wing_loading)

        airspeeds    = np.concatenate((airspeeds[:(Num_of_points + 3)], [V_inters], airspeeds[(Num_of_points + 3):]))
        load_factors = np.concatenate((load_factors[:(Num_of_points + 3)], [load_factor_inters], load_factors[(Num_of_points + 3):]))

        # Pack
        if sign_flag == 1:
            V_n_data.airspeeds.positive          = airspeeds
            V_n_data.load_factors.positive       = load_factors
            V_n_data.gust_load_factors.positive  = gust_load_factors
            V_n_data.Vb_load_factor.positive     = load_factor_inters
            V_n_data.Vb.positive                 = V_inters           
            
        if sign_flag == 2:
            V_n_data.airspeeds.negative          = airspeeds
            V_n_data.load_factors.negative       = load_factors
            V_n_data.gust_load_factors.negative  = gust_load_factors
            V_n_data.Vb_load_factor.negative     = load_factor_inters
            V_n_data.Vb.negative                 = V_inters
        
        # continue stall lines
        Num_of_points_ext       = 5;
        upper_bound             = Num_of_points+3;
        lower_bound             = Num_of_points+2;
        stall_line(V_n_data, upper_bound, lower_bound, Num_of_points_ext, sign_flag)        
        Num_of_points = Num_of_points + Num_of_points_ext + 1

        # Unpack
        if sign_flag == 1:
            airspeeds        = V_n_data.airspeeds.positive
            load_factors     = V_n_data.load_factors.positive
            lift_coefficient = V_n_data.maximum_lift_coefficient

        elif sign_flag == 2:
            airspeeds        = V_n_data.airspeeds.negative
            load_factors     = V_n_data.load_factors.negative
            lift_coefficient = V_n_data.minimum_lift_coefficient

        
        # insert the cruise speed Vc in the positive load factor line
        if load_factors[1] > 0:
            airspeeds    = np.concatenate((airspeeds[:(Num_of_points + 3)], [Vc], airspeeds[(Num_of_points + 3):]))
            load_factors = np.concatenate((load_factors[:(Num_of_points + 3)], [gust_load_factors[2]], \
                                           load_factors[(Num_of_points + 3):]))

    # Intersection between cruise gust load and maximum load at Vc
    elif abs(gust_load_factors[2]) > abs(limit_load):
        V_inters      = 498 * (load_factors[Num_of_points + 2] - 1) * wing_loading / (Kg * Uref_cruise * CLa)
        airspeeds     = np.concatenate((airspeeds[:(Num_of_points + 3)], [V_inters], airspeeds[(Num_of_points + 3):]))
        load_factors  = np.concatenate((load_factors[:(Num_of_points + 3)], [limit_load], \
                                       load_factors[(Num_of_points + 3):]))
        Num_of_points = Num_of_points + 1

        if load_factors[1] > 0:
            airspeeds    = np.concatenate((airspeeds[:(Num_of_points + 3)], [Vc], airspeeds[(Num_of_points + 3):]))
            load_factors = np.concatenate((load_factors[:(Num_of_points + 3)], [gust_load_factors[2]], \
                                           load_factors[(Num_of_points + 3):]))
        else:
            load_factors[len(airspeeds)-2] = gust_load_factors[2]

        
    # Dive speed Gust loads Vd
    gust_load_factors[3] = 1 + Kg * CLa * Vd * Uref_dive / (498 * wing_loading)

    # Resolve the upper half of the dive section
    if load_factors[1] > 0: 
        if abs(gust_load_factors[2]) > abs(limit_load):   
            if abs(gust_load_factors[3]) > abs(limit_load):
                load_factors[len(load_factors) - 2] = gust_load_factors[3]               
            else:
                airspeeds, load_factors = gust_dive_speed_intersection(category_tag, load_factors, gust_load_factors, airspeeds, \
                                                                       len(load_factors)-2, Vc, Vd)

    # Resolve the lower half of the dive section
    else:
        if gust_load_factors[3] < load_factors[len(load_factors) - 1]:
            airspeeds    = np.concatenate((airspeeds[:(len(load_factors) - 1)], [airspeeds[(len(load_factors) - 1)]], \
                                           airspeeds[(len(load_factors) - 1):]))
            load_factors = np.concatenate((load_factors[:(len(load_factors) - 1)], [gust_load_factors[3]], \
                                           load_factors[(len(load_factors) - 1):]))
            
            if abs(gust_load_factors[2]) > abs(limit_load):
                load_factors[len(load_factors) - 2] = gust_load_factors[3]
            else:
                airspeeds, load_factors = gust_dive_speed_intersection(category_tag, load_factors, gust_load_factors, airspeeds, \
                                                                       len(load_factors)-2, Vc, Vd)

            V_n_data.limit_loads.dive.negative = load_factors[len(load_factors) - 2]
        else:
            V_n_data.limit_loads.dive.negative = load_factors[len(load_factors) - 1]

    # gusts load extension for gust lines at Vd
    gust_load_factors = np.append(gust_load_factors, 1 + Kg * CLa * (1.05 * Vd) * Uref_cruise/(498 * wing_loading))
    gust_load_factors = np.append(gust_load_factors, 1 + Kg * CLa * (1.05 * Vd) * Uref_dive/(498 * wing_loading))

    # guts load extension for gust lines at Vb and Vc for a rough gust
    if category_tag == 'commuter':
        gust_load_factors = np.append(gust_load_factors, 1 + Kg * CLa * (1.05 * Vd) * Uref_rough/(498 * wing_loading))
    else:
        gust_load_factors = np.append(gust_load_factors, 0)
      
    # Pack
    if sign_flag == 1:
        V_n_data.airspeeds.positive          = airspeeds
        V_n_data.load_factors.positive       = load_factors
        V_n_data.gust_load_factors.positive  = gust_load_factors
        V_n_data.limit_loads.dive.positive   = load_factors[len(load_factors) - 2]
        
    if sign_flag == 2:
        V_n_data.airspeeds.negative          = airspeeds
        V_n_data.load_factors.negative       = load_factors
        V_n_data.gust_load_factors.negative  = gust_load_factors
    
    return 
#------------------------------------------------------------------------------------------------------------------------

def gust_dive_speed_intersection(category_tag, load_factors, gust_load_factors, airspeeds, element_num, Vc, Vd):

    """ Calculates intersection between the general V-n diagram and the gust load for Vd

    Source:
    S. Gudmundsson "General Aviation Aircraft Design: Applied Methods and Procedures", Butterworth-Heinemann; 1 edition

    Inputs:
    load_factors                                [Unitless]
    gust_load_factors                           [Unitless]
    airspeeds                                   [kts]
    Vc                                          [kts]
    Vd                                          [kts]
    element_num                                 [Unitless]
    
    Outputs:
    airspeeds                                    [kts]
    load_factors                                 [Unitless]

    Properties Used:
    N/A

    Description:
    A specific function for CFR FAR Part 25 regulations. For negative loads, the general curve must go linear from a specific
    load factor at Vc to zero at Vd. Gust loads may put a non-zero load factor at Vd, so an intersection between the old and
    the new curves is required
    """
    
    if load_factors[1] > 0:
        V_inters = (load_factors[element_num] - gust_load_factors[2]) * (Vd - Vc) \
                    / (gust_load_factors[3] - gust_load_factors[2]) + Vc
        load     = load_factors[element_num]
    else:
        if category_tag == 'acrobatic':          
            V_inters = ((gust_load_factors[3] + 1) * Vc + (min(load_factors) - gust_load_factors[2]) * Vd) / \
                       (gust_load_factors[3] - gust_load_factors[2] + min(load_factors) + 1)
            load     = (min(load_factors) + 1) / (Vc - Vd) * V_inters - ((min(load_factors) + 1) / (Vc - Vd) * Vd + 1)
        else:
            V_inters = (gust_load_factors[3] * Vc + (min(load_factors) - gust_load_factors[2]) * Vd) / \
                       (gust_load_factors[3] - gust_load_factors[2] + min(load_factors))
            load     = min(load_factors) * (V_inters - Vd) / (Vc - Vd)


    airspeeds    = np.concatenate((airspeeds[:(element_num)], [V_inters], airspeeds[(element_num):]))
    load_factors = np.concatenate((load_factors[:(element_num)], [load], \
                                        load_factors[(element_num):]))    

    return airspeeds, load_factors
#--------------------------------------------------------------------------------------------------------------------------

def post_processing(category_tag, Uref_rough, Uref_cruise, Uref_dive, V_n_data, vehicle):

    """ Plot graph, save the final figure, and create results output file

    Source:

    Inputs:
    V_n_data.
        airspeeds.positive                  [kts]
        airspeeds.negative                  [kts]
        Vc                                  [kts]
        Vd                                  [kts]
        Vs1.positive                        [kts]
            negative                        [kts]
        Va.positive                         [kts]
            negative                        [kts]
        load_factors.positive               [Unitless]
            negative                        [Unitless]
        gust_load_factors.positive          [Unitless]
            negative                        [Unitless]
        weight                              [lb]
        altitude                            [ft]
    vehicle._base.tag                       [Unitless]
    Uref_rough                              [ft/s]
    Uref_cruise                             [ft/s]
    Uref_dive                               [ft/s]

    Outputs:

    Properties Used:
    N/A

    Description:
    """

    # Unpack
    load_factors_pos        = V_n_data.load_factors.positive
    load_factors_neg        = V_n_data.load_factors.negative
    airspeeds_pos           = V_n_data.airspeeds.positive
    airspeeds_neg           = V_n_data.airspeeds.negative
    Vc                      = V_n_data.Vc
    Vd                      = V_n_data.Vd
    Vs1_pos                 = V_n_data.Vs1.positive
    Vs1_neg                 = V_n_data.Vs1.negative
    Va_pos                  = V_n_data.Va.positive
    Va_neg                  = V_n_data.Va.negative
    gust_load_factors_pos   = V_n_data.gust_load_factors.positive
    gust_load_factors_neg   = V_n_data.gust_load_factors.negative
    weight                  = V_n_data.weight
    altitude                = V_n_data.altitude

    #-----------------------------
    # Plotting the V-n diagram
    #-----------------------------
    fig, ax = plt.subplots()
    ax.fill(airspeeds_pos, load_factors_pos, c='b', alpha=0.3)
    ax.fill(airspeeds_neg, load_factors_neg, c='b', alpha=0.3)
    ax.plot(airspeeds_pos, load_factors_pos, c='b')
    ax.plot(airspeeds_neg, load_factors_neg, c='b')

    # Plotting gust lines
    ax.plot([0, Vc,1.05*Vd],[1,gust_load_factors_pos[2],gust_load_factors_pos[len(gust_load_factors_pos)-3]],'--', c='r', label = ('Gust ' + str(round(Uref_cruise)) + 'fps'))
    ax.plot([0, Vd,1.05*Vd],[1,gust_load_factors_pos[3],gust_load_factors_pos[len(gust_load_factors_pos)-2]],'--', c='g', label = ('Gust ' + str(round(Uref_dive)) + 'fps'))
    ax.plot([0, Vc,1.05*Vd],[1,gust_load_factors_neg[2],gust_load_factors_neg[len(gust_load_factors_neg)-3]],'--', c='r')
    ax.plot([0, Vd,1.05*Vd],[1,gust_load_factors_neg[3],gust_load_factors_neg[len(gust_load_factors_neg)-2]],'--', c='g')

    if category_tag == 'commuter':
        ax.plot([0, 1.05*Vd],[1,gust_load_factors_pos[len(gust_load_factors_pos)-1]],'--', c='m', label = ('Gust ' + str(round(Uref_rough)) + 'fps'))
        ax.plot([0, 1.05*Vd],[1,gust_load_factors_neg[len(gust_load_factors_neg)-1]],'--', c='m')

    # Formating the plot
    ax.set_xlabel('Airspeed, KEAS')
    ax.set_ylabel('Load Factor')
    ax.set_title(vehicle.tag + '  Weight=' + str(round(weight)) + 'lb  ' + ' Altitude=' + str(round(altitude)) + 'ft ')
    ax.legend()
    ax.grid() 

    #---------------------------------
    # Creating results output file
    #---------------------------------
    fres = open("V_n_diagram_results_" + vehicle.tag +".dat","w")
    fres.write('V-n diagram summary\n')
    fres.write('-------------------\n')
    fres.write('Aircraft: ' + vehicle.tag + '\n')
    fres.write('category: ' + vehicle.flight_envelope.category + '\n')
    fres.write('FAR certification: Part ' +  vehicle.flight_envelope.FAR_part_number  + '\n')
    fres.write('Weight = ' + str(round(weight)) + ' lb\n')
    fres.write('Altitude = ' + str(round(altitude)) + ' ft\n')
    fres.write('---------------------------------------------------------------\n\n')
    fres.write('Airspeeds: \n')
    fres.write('    Positive stall speed (Vs1)   = ' + str(round(Vs1_pos,1)) + ' KEAS\n')
    fres.write('    Negative stall speed (Vs1)   = ' + str(round(Vs1_neg,1)) + ' KEAS\n')
    fres.write('    Positive maneuver speed (Va) = ' + str(round(Va_pos,1))  + ' KEAS\n')
    fres.write('    Negative maneuver speed (Va) = ' + str(round(Va_neg,1))  + ' KEAS\n')
    fres.write('    Cruise speed (Vc)            = ' + str(round(Vc,1))      + ' KEAS\n')
    fres.write('    Dive speed (Vd)              = ' + str(round(Vd,1))      + ' KEAS\n')
    fres.write('Load factors: \n')
    fres.write('    Positive limit load factor (n+) = ' + str(round(max(load_factors_pos),2)) + '\n')
    fres.write('    Negative limit load factor (n-) = ' + str(round(min(load_factors_neg),2)) + '\n')
    fres.write('    Positive load factor at Vd      = ' + str(round(V_n_data.limit_loads.dive.positive,2)) + '\n')
    fres.write('    Negative load factor at Vd      = ' + str(round(V_n_data.limit_loads.dive.negative,2)) + '\n')
   
    return
#------------------------------------------------------------------------------------------------------------------------

def convert_keas(V_n_data):

    """ Convert speed to KEAS

    Source:

    Inputs:
    V_n_data.
        airspeeds.positive              [ft/s]
            negative                    [ft/s]
        Vc                              [ft/s]
        Va.positive                     [ft/s]
        Va.negative                     [ft/s]
        Vs1.negative                    [ft/s]
        Vs1.positive                    [ft/s]
        density_ratio                   [Unitless]

    Outputs:
    V_n_data.              
        airspeeds.positive              [kts]
            negative                    [kts]
        Vc                              [kts]
        Va.positive                     [kts]
        Va.negative                     [kts]
        Vs1.negative                    [kts]
        Vs1.positive                    [kts]

    Properties Used:
    N/A

    Description:
    """

    # Unpack
    airspeeds_pos = V_n_data.airspeeds.positive
    airspeeds_neg = V_n_data.airspeeds.negative
    density_ratio = V_n_data.density_ratio
    Vc            = V_n_data.Vc
    Va_pos        = V_n_data.Va.positive
    Va_neg        = V_n_data.Va.negative
    Vs1_neg       = V_n_data.Vs1.negative
    Vs1_pos       = V_n_data.Vs1.positive
    
    airspeeds_pos = airspeeds_pos * Units['ft/s'] / Units.knots * density_ratio
    airspeeds_neg = airspeeds_neg * Units['ft/s'] / Units.knots * density_ratio

    Vs1_pos       = Vs1_pos * Units['ft/s'] / Units.knots * density_ratio
    Vs1_neg       = Vs1_neg * Units['ft/s'] / Units.knots * density_ratio
    Va_pos        = Va_pos * Units['ft/s'] / Units.knots * density_ratio
    Va_neg        = Va_neg * Units['ft/s'] / Units.knots * density_ratio
    Vc            = Vc[0] * Units['ft/s'] / Units.knots * density_ratio
    Vc            = Vc[0]
    
    # Pack
    V_n_data.airspeeds.positive = airspeeds_pos
    V_n_data.airspeeds.negative = airspeeds_neg
    V_n_data.Vc                 = Vc
    V_n_data.Va.positive        = Va_pos
    V_n_data.Va.negative        = Va_neg
    V_n_data.Vs1.negative       = Vs1_neg
    V_n_data.Vs1.positive       = Vs1_pos
    
    
    return


# RCAIDE/Library/Compoments/Fuselage/Segment.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from RCAIDE.Framework.Core import Data, Container
from RCAIDE.Library.Components import Component  

# ---------------------------------------------------------------------------------------------------------------------- 
#  Segment
# ----------------------------------------------------------------------------------------------------------------------   
class Segment(Component):
    """
    A component representing a generic super-ellipse cross-sectional segment of a fuselage. Segments are used 
    to define the shape and dimensions of the fuselage through a series of connected cross-sections.

    Attributes
    ----------
    tag : str
        Identifier for the segment, defaults to 'segment'
        
    prev : Component
        Link to the previous segment in the fuselage chain, defaults to None
        
    next : Component
        Link to the next segment in the fuselage chain, defaults to None
        
    percent_x_location : float
        Longitudinal position as percentage of fuselage length, defaults to 0
        
    percent_y_location : float
        Lateral position as percentage of fuselage width, defaults to 0
        
    percent_z_location : float
        Vertical position as percentage of fuselage height, defaults to 0
        
    height : float
        Vertical dimension of the segment cross-section, defaults to 0
        
    width : float
        Lateral dimension of the segment cross-section, defaults to 0
        
    curvature : float
        Shape parameter controlling cross-section corner rounding, defaults to 2

    Notes
    -----
    Segments are used to build up the complete fuselage geometry through a series of 
    cross-sections. Each segment's position is defined as a percentage of the overall 
    fuselage dimensions, allowing for flexible scaling and positioning.

    **Major Assumptions**
    
    * Segments are connected in sequence to form a continuous surface
    * Cross-sections are symmetric about the vertical plane
    * Transitions between segments are smooth
    
    **Definitions**

    'Cross-section'
        The 2D shape formed by intersecting the fuselage with a plane perpendicular 
        to its longitudinal axis
        
    'Curvature'
        Parameter controlling the smoothness of transition between vertical and 
        horizontal surfaces at the corners of the cross-section

    See Also
    --------
    RCAIDE.Library.Components.Fuselages.Fuselage
        Parent container for fuselage segments
    """

    def __defaults__(self): 
        """
        Sets default values for the fuselage segment attributes.
        """
        self.tag                     = 'segment'
        self.prev                    = None
        self.next                    = None    
        self.percent_x_location      = 0  
        self.percent_y_location      = 0
        self.percent_z_location      = 0 
        self.height                  = 0 
        self.width                   = 0 
        self.curvature               = 2
        self.radius                  = 0

        self.vsp_data                = Data()
        self.vsp_data.xsec_id        = ''       
        self.vsp_data.shape          = ''        
         
class Segment_Container(Container):
    """
    Container class for managing fuselage segments. Provides organization and 
    access methods for segment components.

    Notes
    -----
    This container is designed to hold and manage Segment objects that define 
    the fuselage shape. It inherits from the base Container class and provides 
    specialized functionality for fuselage segments.

    See Also
    --------
    RCAIDE.Framework.Core.Container
        Base container class providing core functionality
    RCAIDE.Library.Components.Fuselages.Segments.Segment
        The segment components stored in this container
    """     

    def get_children(self):
        """
        Returns a list of allowable child component types for the segment container.

        Returns
        -------
        list
            Empty list as segments do not contain child components
        """       
        return []


# RCAIDE/Framework/Analyses/Common/Process_Geometry.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports  
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Analyses import Process

# ----------------------------------------------------------------------------------------------------------------------
#  Analysis
# ----------------------------------------------------------------------------------------------------------------------
class Process_Geometry(Process):
    """A process for evaluate over a component group.

    Assumptions:
    None

    Source:
    N/A
    """      
    
    geometry_key = None
    
    def __init__(self,geometry_key):
        """Sets the geometry key for this process.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        geometry_key      <string>

        Outputs:
        None

        Properties Used:
        self.geometry_key <string>
        """          
        self.geometry_key = geometry_key
    
    def evaluate(self,state,settings,geometry):
        """Evaluates preset processes for each component.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        state     (passed to an evaluation function)
        setting   (passed to an evaluation function)
        geometry  (used to get keys and passed to an evaluation function)

        Outputs:
        None

        Properties Used:
        self.geometry_key <string>
        """             
        geometry_items = geometry.deep_get(self.geometry_key)
        
        results = Data()
        
        for key, this_geometry in geometry_items.items():
            result = Process.evaluate(self,state,settings,this_geometry)
            results[key] = result
            
        return results
        
        
        
        


# RCAIDE/Library/Methods/Geometry/Two_Dimensional/Airfoil/generate_interpolated_airfoils.py
# 
# 
# Created:  Jul 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    

from RCAIDE.Library.Methods.Geometry.Airfoil.import_airfoil_geometry import import_airfoil_geometry 
import numpy as np
import os

# ----------------------------------------------------------------------------------------------------------------------
#  generate_interpolated_airfoils
# ----------------------------------------------------------------------------------------------------------------------    
def generate_interpolated_airfoils(a1, a2, nairfoils, npoints=200, save_filename="Transition"):
    """ Takes in two airfoils, interpolates between their coordinates to generate new
    airfoil geometries and saves new airfoil files.
    
    Assumptions: Linear geometric transition between airfoils
    
    Source: None
    
    Inputs:
    a1                 first airfoil                                [ airfoil ]
    a2                 second airfoil                               [ airfoil ]
    nairfoils          number of airfoils                           [ unitless ]
    
    """
    
    # import airfoil geometry for the two airfoils 
    a1_name           = os.path.basename(a1)
    a2_name           = os.path.basename(a2)
    a_geo_1           = import_airfoil_geometry(a1,npoints)
    a_geo_2           = import_airfoil_geometry(a2,npoints)
    
    # for each point around the airfoil, interpolate between the two given airfoil coordinates
    z = np.linspace(0,1,nairfoils)
    
    y_u_lb = a_geo_1.y_upper_surface 
    y_u_ub = a_geo_2.y_upper_surface 
    y_l_lb = a_geo_1.y_lower_surface 
    y_l_ub = a_geo_2.y_lower_surface      
     
    x_u_lb = a_geo_1.x_upper_surface 
    x_u_ub = a_geo_2.x_upper_surface 
    x_l_lb = a_geo_1.x_lower_surface 
    x_l_ub = a_geo_2.x_lower_surface     
    
    # broadcasting interpolation
    y_n_upper = (z[None,...] * (y_u_ub[...,None] - y_u_lb[...,None]) + (y_u_lb[...,None])).T
    y_n_lower = (z[None,...] * (y_l_ub[...,None] - y_l_lb[...,None]) + (y_l_lb[...,None])).T
    x_n_upper = (z[None,...] * (x_u_ub[...,None] - x_u_lb[...,None]) + (x_u_lb[...,None])).T
    x_n_lower = (z[None,...] * (x_l_ub[...,None] - x_l_lb[...,None]) + (x_l_lb[...,None])).T
    
    
    # save new airfoil geometry files:
    new_files = {'a_{}'.format(i+1): [] for i in range(nairfoils-2)}
    airfoil_files = []

    for k in range(nairfoils-2):
        # create new files and write title block for each new airfoil
        title_block     = "Airfoil Transition "+str(k+1)+" between "+a1_name+" and "+a2_name+"\n 61. 61.\n\n"
        file            ='a_'+str(k+1)
        new_files[file] = open(save_filename + str(k+1) +".txt", "w+")
        new_files[file].write(title_block)
        
        y_n_u = np.reshape(y_n_upper[k+1],(npoints//2,1))
        y_n_l = np.reshape(y_n_lower[k+1],(npoints//2,1))
        x_n_u = np.reshape(x_n_upper[k+1],(npoints//2,1))
        x_n_l = np.reshape(x_n_lower[k+1],(npoints//2,1))
        
        airfoil_files.append(new_files[file].name)
        upper_data = np.append(x_n_u, y_n_u,axis=1)
        lower_data = np.append(x_n_l, y_n_l,axis=1)

        # write lines to files
        for lines in upper_data: #upper_data[file]:
            line = str(lines[0]) + " " + str(lines[1]) + "\n"
            new_files[file].write(line)
            
        new_files[file].write("\n")
        
        for lines in lower_data: #[file]:
            line = str(lines[0]) + " " + str(lines[1]) + "\n"
            new_files[file].write(line)
            
        new_files[file].close()
        
    # plot new and original airfoils:
    airfoil_files.insert(0,a1)
    airfoil_files.append(a2)
     
    return airfoil_files



# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/unpack_electric_ducted_fan_unknowns.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke   

# ---------------------------------------------------------------------------------------------------------------------- 
#  unpack electric ducted_fan network unknowns 
# ----------------------------------------------------------------------------------------------------------------------  

def unpack_electric_ducted_fan_unknowns(propulsor,segment): 
    results = segment.state.conditions.energy[propulsor.tag]
    motor   = propulsor.motor  
    results[motor.tag].rotor_power_coefficient = segment.state.unknowns[propulsor.tag  + '_ducted_fan_cp'] 
    return 


# RCAIDE/Methods/Energy/Sources/Battery/Lithium_Ion_LFP/update_lfp_cell_age.py
# 
# 
# Created: Nov 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np  
 
# ----------------------------------------------------------------------------------------------------------------------
# update_lfp_cell_age
# ----------------------------------------------------------------------------------------------------------------------  
def update_lfp_cell_age(battery_module,segment, battery_conditions,increment_battery_age_by_one_day):  
    """ This is an aging model for 26650 A123 LFP cell. 
   
    Source: 
    Nájera, J., J.R. Arribas, R.M. De Castro, and C.S. Núñez. 
    “Semi-Empirical Ageing Model for LFP and NMC Li-Ion Battery Chemistries.”
    Journal of Energy Storage 72 (November 2023): 108016.
    https://doi.org/10.1016/j.est.2023.108016.

      
    Assumptions:
    None
    """
    SOC                = battery_conditions.cell.state_of_charge
    I                  = battery_conditions.cell.current
    t                  = battery_conditions.cell.cycle_in_day         
    charge_thougput    = battery_conditions.cell.charge_throughput
    Temp               = (battery_conditions.cell.temperature) 
    C_rate             = np.sqrt(np.mean(I**2)) /battery_module.cell.nominal_capacity
    
    # Semi Emperical aging model  
    E_fade_factor = 1-(((Temp**2*2.0916e-8)+(-1.2179e-5*Temp)+0.0018)*np.exp(((-1.7082e-6*Temp)+0.0556)*C_rate) \
                        * charge_thougput + (5.9808e6) * np.exp(0.68989*SOC) * np.exp(-6.4647e3/Temp) * t**(0.5))


    battery_conditions.cell.capacity_fade_factor     = np.minimum(E_fade_factor[-1],battery_conditions.cell.capacity_fade_factor)
    
    if increment_battery_age_by_one_day:
        battery_conditions.cell.cycle_in_day += 1 # update battery age by one day 
    
    return 



# RCAIDE/Library/Plots/Thermal_Management/plot_wavy_channel_conditions.py
# 
# 
# Created:  Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#   plot_wavy_channel_conditions
# ----------------------------------------------------------------------------------------------------------------------   
def plot_wavy_channel_conditions(wavy_channel, results, coolant_line, save_figure,show_legend ,save_filename,file_type , width, height):
    """Plots the Wavy Channel Heat Acqusition conditions throughout flight.
    
     Assumptions:
     None
    
     Source:
     None
    
     Inputs:
     results.segments.conditions.energy[coolant_line.tag][wavy_channel.tag].
                                                                       outlet_coolant_temperature
                                                                       coolant_mass_flow_rate
                                                                       power
                                                                
                                                                       
     Outputs:
     Plots
    
     Properties Used:
     N/A	
     """ 
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))
    
    fig = plt.figure('Identical_'+ save_filename)
    fig.set_size_inches(width,height) 
    axis_1 = plt.subplot(2,2,1)
    axis_2 = plt.subplot(2,2,2) 
    axis_3 = plt.subplot(2,2,3)

    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus in busses:
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)): 
                        time                            = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min    
                        wavy_channel_conditions         = results.segments[i].conditions.energy[coolant_line.tag][wavy_channel.tag]   
                        outlet_coolant_temperature      = wavy_channel_conditions.outlet_coolant_temperature[:,0]
                        coolant_mass_flow_rate          = wavy_channel_conditions.coolant_mass_flow_rate[:,0]
                        power                           = wavy_channel_conditions.power[:,0]       
                
                        if i == 0:                
                            axis_1.plot(time, outlet_coolant_temperature, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width, label = wavy_channel.tag)
                        else:
                            axis_1.plot(time, outlet_coolant_temperature, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_1.set_ylabel(r'Coolant Temp. (K)') 
                        set_axes(axis_1)     
                         
                        axis_2.plot(time, coolant_mass_flow_rate, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_2.set_ylabel(r'Coolant $\dot{m}$ (kg/s)')
                        set_axes(axis_2) 
                 
                        axis_3.plot(time, power, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_3.set_ylabel(r'HAS Power (W)')
                        axis_3.set_xlabel(r'Time (mins)')
                        set_axes(axis_3)
        
    if show_legend:          
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})     
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text   = 'Wavy_Channel_Properties'       
    fig.suptitle(title_text) 
    
    if save_figure:
        plt.savefig(wavy_channel.tag + file_type)    
    return fig 


# dynamo_efficiency.py
#
# Created:  Feb 2022,  S. Claridge

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE imports 
import numpy as np 

# ----------------------------------------------------------------------
#  Estimated efficiency of HTS Dynamo 
# ---------------------------------------------------------------------- 
def efficiency_curve(Dynamo, current):

    """ This sets the default values.

    Assumptions:
        The efficiency curve of the Dynamo is a parabola 

    Source:
        "Practical Estimation of HTS Dynamo Losses" - Kent Hamilton, Member, IEEE, Ratu Mataira-Cole, Jianzhao Geng, Chris Bumby, Dale Carnegie, and Rod Badcock, Senior Member, IEEE

    Inputs:
        current        [A]

    Outputs:
        efficiency      [W/W]

    Properties Used:
        None
    """     

    x = np.array(current)

    if np.any(x > Dynamo.rated_current * 1.8 ) or np.any(x < Dynamo.rated_current * 0.2): #Plus minus 80
        print("Current out of range")
        return 0 

    a          = ( Dynamo.efficiency ) / np.square(Dynamo.rated_current) #one point on the graph is assumed to be  (0, 2 * current), 0  = a (current ^ 2) + efficiency 
    
    efficiency = -a * (np.square( x - Dynamo.rated_current) ) +  Dynamo.efficiency # y = -a(x - current)^2 + efficieny 

    return   efficiency


# RCAIDE/Methods/Aerodynamics/Common/Drag/wave_drag.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports   
import numpy as np
from RCAIDE.Library.Components.Wings import Main_Wing

# ----------------------------------------------------------------------
#  Wave Drag 
# ---------------------------------------------------------------------- 
def wave_drag(conditions,wing):
    """Computes wave drag due to lift

    Assumptions:
    Main wing is the primary lift contributor

    Source:
    Yoshida, Kenji. "Supersonic drag reduction technology in the scaled supersonic 
    experimental airplane project by JAXA."

    Args:
    conditions.freestream.mach_number        [Unitless]
    conditions.aerodynamics.coefficients.lift [Unitless]
    wing.spans.projected                     [m]
    wing.total_length                        [m]
    wing.aspect_ratio                        [-]

    Returns:
    wave_drag_lift                           [Unitless] 
    """  

    # Unpack
    freestream  = conditions.freestream 
    Mach        = freestream.mach_number * 1.0
    
    # Lift coefficient 
    if isinstance(wing,Main_Wing):
        CL = conditions.aerodynamics.coefficients.lift.total 
        l  = np.maximum(wing.total_length,wing.chords.root)        
    else:
        CL = np.zeros_like(conditions.aerodynamics.coefficients.lift.total)
        l  = np.maximum(wing.total_length,wing.chords.root) 

    # JAXA method
    s    = wing.spans.projected / 2
    AR   = wing.aspect_ratio
    p    = 2/AR*s/l
    beta = np.sqrt(Mach[Mach >= 1.01]**2-1)
    
    Kw = (1+1/p)*fw(beta*s/l)/(2*beta**2*(s/l)**2)
    
    # Ignore area comparison since this is full vehicle CL
    CDwl           = CL[Mach >= 1.01]**2 * (beta**2/np.pi*p*(s/l)*Kw)
    wave_drag_lift = np.zeros_like(Mach)
    wave_drag_lift[Mach >= 1.01] = CDwl

    return wave_drag_lift

def fw(x):
    """Helper function for lift wave drag computations.

    Assumptions:
    N/A

    Source:
    Yoshida, Kenji. "Supersonic drag reduction technology in the scaled supersonic 
    experimental airplane project by JAXA."

    Args:
    x    [Unitless]

    Returns:
    ret  [Unitless]

    Properties Used:
    N/A
    """  
    
    ret = np.zeros_like(x)
    
    ret[x > 0.178] = 0.4935 - 0.2382*x[x > 0.178] + 1.6306*x[x > 0.178]**2 - \
        0.86*x[x > 0.178]**3 + 0.2232*x[x > 0.178]**4 - 0.0365*x[x > 0.178]**5 - 0.5
    
    return ret


# RCAIDE/Library/Missions/Segments/Ground/Takeoff.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports  
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
# unpack unknowns
# ---------------------------------------------------------------------------------------------------------------------- 
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Builds on the initialize conditions for common

    Source:
    N/A

    Inputs:
    segment.throttle                           [unitless]
    conditions.frames.inertial.position_vector [meters]
    conditions.weights.total_mass              [kilogram]

    Outputs:
    conditions.weights.total_mass              [kilogram]
    conditions.frames.inertial.position_vector [unitless]
    conditions.propulsion.throttle             [meters]
    
    Properties Used:
    N/A
    """  

    # use the common initialization # unpack inputs
    alt      = segment.altitude 
    v0       = segment.velocity_start
    vf       = segment.velocity_end 
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]   

    if v0  is None: 
        v0 = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # avoid having zero velocity since aero and propulsion models need non-zero Reynolds number
    if v0 == 0.0: v0 = 0.01
    if vf == 0.0: vf = 0.01
    
    # intial and final speed cannot be the same
    if v0 == vf:
        vf = vf + 0.01
        
    # repack
    segment.air_speed_start = v0
    segment.air_speed_end   = vf
    
    initialized_velocity = (vf - v0)*segment.state.numerics.dimensionless.control_points + v0
    
    # Initialize the x velocity unknowns to speed convergence:
    segment.state.unknowns.ground_velocity = initialized_velocity[1:,0]    

    # pack conditions 
    conditions = segment.state.conditions    
    conditions.frames.inertial.velocity_vector[:,0] = initialized_velocity[:,0]
    conditions.ground.incline[:,0]                  = segment.ground_incline
    conditions.ground.friction_coefficient[:,0]     = segment.friction_coefficient   
    conditions.freestream.altitude[:,0]             = alt
    conditions.frames.inertial.position_vector[:,2] = -alt   
    conditions.weights.total_mass[:,0]              = segment.analyses.weights.vehicle.mass_properties.takeoff
    conditions.frames.inertial.position_vector[:,:] = conditions.frames.inertial.position_vector[0,:][None,:][:,:]


# RCAIDE/Library/Plots/Geometry/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .Common            import * 
from .plot_3d_vehicle                  import plot_3d_vehicle
from .plot_3d_vehicle                  import plot_3d_energy_network 
from .plot_3d_vehicle                  import generate_3d_vehicle_geometry_data 
from .plot_3d_rotor                    import plot_3d_rotor
from .plot_3d_rotor                    import generate_3d_blade_points 
from .plot_3d_nacelle                  import plot_3d_nacelle 
from .plot_3d_nacelle                  import generate_3d_basic_nacelle_points
from .plot_3d_nacelle                  import generate_3d_BOR_nacelle_points
from .plot_3d_nacelle                  import generate_3d_stack_nacelle_points
from .plot_3d_wing                     import plot_3d_wing
from .plot_3d_wing                     import generate_3d_wing_points
from .plot_3d_vehicle_vlm_panelization import plot_3d_vehicle_vlm_panelization
from .plot_airfoil                     import plot_airfoil
from .plot_rotor                       import plot_rotor 


# RCAIDE/Library/Components/Energy/Sources/Fuel_Tanks/Central_Fuel_Tank.py
# 
# 
# Created:  September 2024, A. Molloy and M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from .Fuel_Tank  import Fuel_Tank 
from RCAIDE.Library.Methods.Energy.Sources.Fuel_Tanks.append_fuel_tank_conditions import append_fuel_tank_conditions 

# ----------------------------------------------------------------------------------------------------------------------
#  Fuel Tank
# ---------------------------------------------------------------------------------------------------------------------    
class Central_Fuel_Tank(Fuel_Tank):
    """
    Class for modeling central fuel tank characteristics and behavior
    
    Attributes
    ----------
    tag : str
        Identifier for the fuel tank (default: 'Central_fuel_tank')
        
    fuel_selector_ratio : float
        Ratio of fuel flow allocation (default: 1.0)
        
    mass_properties.empty_mass : float
        Mass of empty tank structure [kg] (default: 0.0)
        
    secondary_fuel_flow : float
        Secondary fuel flow rate [kg/s] (default: 0.0)
        
    length : float
        Tank length [m] (default: 0.0)
        
    width : float
        Tank width [m] (default: 0.0)
        
    height : float
        Tank height [m] (default: 0.0)
        
    fuel : Component, optional
        Fuel type stored in tank (default: None)

    Notes
    -----
    The central fuel tank is located in the aircraft's center section,
    often integrated with the wing box or fuselage structure. 

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks.Fuel_Tank
        Base fuel tank class
    RCAIDE.Library.Components.Energy.Sources.Fuel_Tanks.Wing_Fuel_Tank
        Wing-mounted fuel tank class
    """
    
    def __defaults__(self):
        """
        Sets default values for central fuel tank attributes
        """          
        self.tag                         = 'Central_fuel_tank'
        self.fuel_selector_ratio         = 1.0 
        self.mass_properties.empty_mass  = 0.0   
        self.secondary_fuel_flow         = 0.0
        self.length                      = 0.0
        self.width                       = 0.0
        self.height                      = 0.0
        self.fuel                        = None
         

    def append_operating_conditions(self,segment,fuel_line):  
        """
        Append fuel tank operating conditions for a flight segment
        
        Parameters
        ----------
        segment : Segment
            Flight segment containing state conditions
        fuel_line : Component
            Connected fuel line component
        """
        append_fuel_tank_conditions(self,segment, fuel_line)  
        return                                          


#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core                              import Units 
from RCAIDE.Framework.Mission.Common                    import Results  
from RCAIDE.Framework.Mission.Segments.Segment          import Segment  
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.compute_rotor_performance import compute_rotor_performance 
 
import numpy as np

# ------------------------------------------------------------------------------
#   Propeller Analysis
# ------------------------------------------------------------------------------ 
def propeller_aerodynamic_analysis(propeller,
                           velocity_range,
                           angular_velocity = 2500*Units.rpm,
                           angle_of_attack = 0, 
                           altitude = 0,
                           delta_isa =0 ): 
    # design aircract 
    electric_rotor                 = RCAIDE.Library.Components.Propulsors.Electric_Rotor()  
    electric_rotor.rotor           = propeller 
    
    # operarting states  
    ctrl_pts          =  len(velocity_range) 
    
    # Find the operating conditions
    atmosphere                                             = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere_conditions                                  = atmosphere.compute_values(propeller.cruise.design_altitude)  
 
    segment                                                = Segment()  
    conditions                                             = Results()  
    conditions.aerodynamics.angle_of_attack                = np.array([[angle_of_attack]])
    conditions.freestream.density                          = atmosphere_conditions.density 
    conditions.freestream.dynamic_viscosity                = atmosphere_conditions.dynamic_viscosity 
    conditions.freestream.speed_of_sound                   = atmosphere_conditions.speed_of_sound 
    conditions.freestream.temperature                      = atmosphere_conditions.temperature 
    conditions.frames.planet.true_course                   = np.zeros((ctrl_pts,3,3)) 
    conditions.frames.planet.true_course[:,0,0]            = np.cos(0.0),
    conditions.frames.planet.true_course[:,0,1]            = - np.sin(0.0)
    conditions.frames.planet.true_course[:,1,0]            = np.sin(0.0)
    conditions.frames.planet.true_course[:,1,1]            = np.cos(0.0) 
    conditions.frames.planet.true_course[:,2,2]            = 1 
    conditions.frames.wind.transform_to_inertial           = np.zeros((ctrl_pts,3,3))   
    conditions.frames.wind.transform_to_inertial[:,0,0]    = np.cos(0.0) 
    conditions.frames.wind.transform_to_inertial[:,0,2]    = np.sin(0.0) 
    conditions.frames.wind.transform_to_inertial[:,1,1]    = 1 
    conditions.frames.wind.transform_to_inertial[:,2,0]    = -np.sin(0.0) 
    conditions.frames.wind.transform_to_inertial[:,2,2]    = np.cos(0.0)  
    conditions.frames.body.transform_to_inertial           = np.zeros((ctrl_pts,3,3))
    conditions.frames.body.transform_to_inertial[:,0,0]    = np.cos(angle_of_attack)
    conditions.frames.body.transform_to_inertial[:,0,2]    = np.sin(angle_of_attack)
    conditions.frames.body.transform_to_inertial[:,1,1]    = 1
    conditions.frames.body.transform_to_inertial[:,2,0]    = -np.sin(angle_of_attack)
    conditions.frames.body.transform_to_inertial[:,2,2]    = np.cos(angle_of_attack)  
    segment.state.conditions                               = conditions 
 

    electric_rotor.append_operating_conditions(segment) 
    for tag, item in  electric_rotor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,electric_rotor)
            
    # Run BEMT
    segment.state.conditions.expand_rows(ctrl_pts)

    conditions.frames.inertial.velocity_vector             = velocity_range
    conditions.freestream.mach_number                      = conditions.frames.inertial.velocity_vector / conditions.freestream.speed_of_sound
    
    rotor_conditions             =  segment.state.conditions.energy[electric_rotor.tag][propeller.tag]        
    rotor_conditions.omega       = angular_velocity
    compute_rotor_performance(electric_rotor,segment.state)
     
    results =  segment.state.conditions.energy[electric_rotor.tag][propeller.tag] 
    return  results


# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/create_avl_datastructures.py
#  
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core import  Data ,  Units 
from RCAIDE.Library.Components.Wings.Control_Surfaces                                      import Aileron , Elevator , Slat , Flap , Rudder 
from RCAIDE.Library.Methods.Aerodynamics.Athena_Vortex_Lattice.write_avl_airfoil_file      import write_avl_airfoil_file  
from .AVL_Objects.Wing                                                                     import Wing, Section, Control_Surface
from .AVL_Objects.Body                                                                     import Body

# package imports
import  numpy as  np


# ----------------------------------------------------------------------------------------------------------------------
#  create_avl_datastructures
# ----------------------------------------------------------------------------------------------------------------------  
def translate_avl_wing(rcaide_wing):
    """ Translates wing geometry from the vehicle setup to AVL format

    Assumptions:
        None

    Source:
        None

    Inputs:
        rcaide_wing.tag                                                          [-]
        rcaide_wing.symmetric                                                    [boolean]
        rcaide_wing.verical                                                      [boolean]
        rcaide_wing - passed into the populate_wing_sections function            [data stucture]

    Outputs:
        w - aircraft wing in AVL format                                         [data stucture] 

    Properties Used:
        N/A
    """         
    w                 = Wing()
    w.tag             = rcaide_wing.tag
    w.symmetric       = rcaide_wing.symmetric
    w.vertical        = rcaide_wing.vertical
    w                 = populate_wing_sections(w,rcaide_wing)

    return w

def translate_avl_body(rcaide_body):
    """ Translates body geometry from the vehicle setup to AVL format

    Assumptions:
        None

    Source:
        None

    Inputs:
        body.tag                                                       [-]
        rcaide_wing.lengths.total                                       [meters]    
        rcaide_body.lengths.nose                                        [meters]
        rcaide_body.lengths.tail                                        [meters]
        rcaide_wing.verical                                             [meters]
        rcaide_body.width                                               [meters]
        rcaide_body.heights.maximum                                     [meters]
        rcaide_wing - passed into the populate_body_sections function   [data stucture]

    Outputs:
        b - aircraft body in AVL format                                [data stucture] 

    Properties Used:
        N/A
    """  
    b                 = Body()
    b.tag             = rcaide_body.tag
    b.symmetric       = True
    b.lengths.total   = rcaide_body.lengths.total
    b.lengths.nose    = rcaide_body.lengths.nose
    b.lengths.tail    = rcaide_body.lengths.tail
    b.widths.maximum  = rcaide_body.width
    b.heights.maximum = rcaide_body.heights.maximum
    b                 = populate_body_sections(b,rcaide_body)

    return b

def populate_wing_sections(avl_wing,rcaide_wing): 
    """ Creates sections of wing geometry and populates the AVL wing data structure

    Assumptions:
        None

    Source:
        None

    Inputs:
        avl_wing.symmetric                         [boolean]
        rcaide_wing.spans.projected                 [meters]
        rcaide_wing.origin                          [meters]
        rcaide_wing.dihedral                        [radians]
        rcaide_wing.segments.sweeps.leading_edge    [radians]
        rcaide_wing.segments.root_chord_percent     [-]
        rcaide_wing.segments.percent_span_location  [-]
        rcaide_wing.segments.sweeps.quarter_chord   [radians]
        rcaide_wing.segment.twist                   [radians]

    Outputs:
        avl_wing - aircraft wing in AVL format     [data stucture] 

    Properties Used:
        N/A
    """           
        
    # obtain the geometry for each segment in a loop                                            
    symm                 = avl_wing.symmetric
    semispan             = rcaide_wing.spans.projected*0.5 * (2 - symm)
    avl_wing.semispan    = semispan   
    root_chord           = rcaide_wing.chords.root
    segments             = rcaide_wing.segments
    segment_names        = list(segments.keys())
    n_segments           = len(segment_names) 
    origin               = rcaide_wing.origin  
        
    if n_segments>0: 
        for i_segs in range(n_segments): 
            current_seg =  segment_names[i_segs]
            if (i_segs == n_segments-1):
                sweep = 0                                 
            else: # This converts all sweeps defined by the quarter chord to leading edge sweep since AVL needs the start of each wing section
                # from the leading edge coordinate and not the quarter chord coordinate 
                next_seg    =  segment_names[i_segs+1]
                
                if segments[current_seg].sweeps.leading_edge is not None: 
                    # If leading edge sweep is defined 
                    sweep       = segments[current_seg].sweeps.leading_edge  
                else:   
                    # If quarter chord sweep is defined, convert it to leading edge sweep
                    sweep_quarter_chord = segments[current_seg].sweeps.quarter_chord 
                    chord_fraction      = 0.25                          
                    segment_root_chord  = root_chord*segments[current_seg].root_chord_percent
                    segment_tip_chord   = root_chord*segments[next_seg].root_chord_percent
                    segment_span        = semispan*(segments[next_seg].percent_span_location - segments[current_seg].percent_span_location )
                    sweep               = np.arctan(((segment_root_chord*chord_fraction) + (np.tan(sweep_quarter_chord )*segment_span - chord_fraction*segment_tip_chord)) /segment_span) 
            dihedral       = segments[current_seg].dihedral_outboard   
    
            # append section 
            section        = Section() 
            section.tag    = segments[current_seg].tag
            section.chord  = root_chord*segments[current_seg].root_chord_percent 
            section.twist  = segments[current_seg].twist/Units.degrees    
            section.origin = origin # first origin in wing root, overwritten by section origin
            if isinstance(segments[current_seg].airfoil, RCAIDE.Library.Components.Airfoils.Airfoil):
                if type(segments[current_seg].airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
                    section.naca_airfoil         = segments[current_seg].airfoil.NACA_4_Series_code
                else:
                    section.airfoil_coord_file   = write_avl_airfoil_file(segments[current_seg].airfoil.coordinate_file)
                    
            # append section to wing
            avl_wing.append_section(section)   
    
            if (i_segs == n_segments-1):
                return avl_wing 
            else:
                # condition for the presence of control surfaces in segment 
                if getattr(rcaide_wing,'control_surfaces',False):   
                    root_chord_percent = segments[current_seg].root_chord_percent
                    tip_chord_percent  = segments[next_seg].root_chord_percent
                    tip_percent_span   = segments[next_seg].percent_span_location
                    root_percent_span  = segments[current_seg].percent_span_location
                    root_twist         = segments[current_seg].twist
                    tip_twist          = segments[next_seg].twist             
                    tip_airfoil        = segments[next_seg].airfoil 
                    seg_tag            = segments[next_seg].tag 
                    
                    # append control surfaces
                    append_avl_wing_control_surfaces(rcaide_wing,avl_wing,semispan,root_chord_percent,tip_chord_percent,tip_percent_span,
                                                     root_percent_span,root_twist,tip_twist,tip_airfoil,seg_tag,dihedral,origin,sweep) 
    
            # update origin for next segment 
            segment_percent_span =    segments[next_seg].percent_span_location - segments[current_seg].percent_span_location     
            if avl_wing.vertical:
                inverted_wing = -np.sign(abs(dihedral) - np.pi/2)
                if inverted_wing  == 0:
                    inverted_wing  = 1
                dz = inverted_wing*semispan*segment_percent_span
                dy = dz*np.tan(dihedral)
                l  = dz/np.cos(dihedral)
                dx = l*np.tan(sweep)
            else:
                inverted_wing = np.sign(dihedral)
                if inverted_wing  == 0:
                    inverted_wing  = 1
                dy = inverted_wing*semispan*segment_percent_span
                dz = dy*np.tan(dihedral)
                l  = dy/np.cos(dihedral)
                dx = l*np.tan(sweep)
            origin= [[origin[0][0] + dx , origin[0][1] + dy, origin[0][2] + dz]]  
        
    else:    
        dihedral              = rcaide_wing.dihedral
        if rcaide_wing.sweeps.leading_edge  is not None: 
            sweep      = rcaide_wing.sweeps.leading_edge
        else:  
            sweep_quarter_chord = rcaide_wing.sweeps.quarter_chord 
            chord_fraction      = 0.25                          
            segment_root_chord  = rcaide_wing.chords.root
            segment_tip_chord   = rcaide_wing.chords.tip
            segment_span        = semispan 
            sweep       = np.arctan(((segment_root_chord*chord_fraction) + (np.tan(sweep_quarter_chord )*segment_span - chord_fraction*segment_tip_chord)) /segment_span)  
        avl_wing.semispan     = semispan 
        
        # define root section 
        root_section          = Section()
        root_section.tag      = 'root_section'
        root_section.origin   = origin
        root_section.chord    = rcaide_wing.chords.root 
        root_section.twist    = rcaide_wing.twists.root/Units.degrees
        
        # append control surfaces 
        tip_airfoil       = rcaide_wing.airfoil 
        seg_tag            = 'section'   
        append_avl_wing_control_surfaces(rcaide_wing,avl_wing,semispan,1.0,rcaide_wing.taper,1.0,
                                         0.0,rcaide_wing.twists.root,rcaide_wing.twists.tip,tip_airfoil,seg_tag,dihedral,origin,sweep)  
        
        # define tip section
        tip_section           = Section()
        tip_section.tag       = 'tip_section'
        tip_section.chord     = rcaide_wing.chords.tip 
        tip_section.twist     = rcaide_wing.twists.tip/Units.degrees  

        # assign location of wing tip         
        if avl_wing.vertical:
            tip_section.origin    = [[origin[0][0]+semispan*np.tan(sweep), origin[0][1]+semispan*np.tan(dihedral), origin[0][2]+semispan]]
        else: 
            tip_section.origin    = [[origin[0][0]+semispan*np.tan(sweep), origin[0][1]+semispan,origin[0][2]+semispan*np.tan(dihedral)]]

        # assign wing airfoil
        if  (rcaide_wing.airfoil !=  None) and (isinstance(rcaide_wing.airfoil) == RCAIDE.Library.Components.Airfoils.Airfoil) :
            root_section.airfoil_coord_file  = rcaide_wing.airfoil.coordinate_file          
            tip_section.airfoil_coord_file   = rcaide_wing.airfoil.coordinate_file     

        avl_wing.append_section(root_section)
        avl_wing.append_section(tip_section)
        
    return avl_wing

def append_avl_wing_control_surfaces(rcaide_wing,avl_wing,semispan,root_chord_percent,tip_chord_percent,tip_percent_span,
                                     root_percent_span,root_twist,tip_twist,tip_airfoil,seg_tag,dihedral,origin,sweep):

    """ Converts control surfaces on a rcaide wing to sections in avl wing

    Assumptions:
        None

    Source:
        None

    Inputs: 
        rcaide_wing           [-]
        avl_wing             [-]
        semispan             [meters]
        root_chord_percent   [unitless]
        tip_chord_percent    [unitless]
        tip_percent_span     [unitless]
        root_percent_span    [unitless]
        root_twist           [radians]
        tip_twist            [radians]
        tip_airfoil          [unitless]
        seg_tag              [unitless]
        dihedral             [radians]
        origin               [meters]
        sweep                [radians]
        
    Outputs: 
        None

    Properties Used:
        N/A
    """         

    root_chord    = rcaide_wing.chords.root                    
    section_spans = []
    for cs in rcaide_wing.control_surfaces:     
        # Create a vectorof all the section breaks from control surfaces on wings.
        # Section breaks include beginning and end of control surfaces as well as the end of segment       
        control_surface_start = semispan*cs.span_fraction_start
        control_surface_end   = semispan*cs.span_fraction_end
        if (control_surface_start < semispan*tip_percent_span) and (control_surface_start >= semispan*root_percent_span) : 
            section_spans.append(control_surface_start) 
        if (control_surface_end  < semispan*tip_percent_span) and (control_surface_end  >= semispan*root_percent_span):                         
            section_spans.append(control_surface_end)                                
    ordered_section_spans = sorted(list(set(section_spans)))     # sort the section_spans in order to create sections in spanwise order
    num_sections = len(ordered_section_spans)                    # count the number of sections breaks that the segment will contain    \

    for section_count in range(num_sections):        
        # create and append sections onto avl wing structure  
        if ordered_section_spans[section_count] == semispan*root_percent_span:  
            # if control surface begins at beginning of segment, redundant section is removed
            section_tags = list(avl_wing.sections.keys())
            del avl_wing.sections[section_tags[-1]]

        # create section for each break in the wing        
        section                   = Section()              
        section.tag               = seg_tag + '_section_'+ str(ordered_section_spans[section_count]) + 'm'
        root_section_chord        = root_chord*root_chord_percent
        tip_section_chord         = root_chord*tip_chord_percent
        semispan_section_fraction = (ordered_section_spans[section_count] - semispan*root_percent_span)/(semispan*(tip_percent_span - root_percent_span ))   
        section.chord             = np.interp(semispan_section_fraction,[0.,1.],[root_section_chord,tip_section_chord])
        root_section_twist        = root_twist/Units.degrees 
        tip_section_twist         = root_chord*tip_twist/Units.degrees  
        section.twist             = np.interp(semispan_section_fraction,[0.,1.],[root_section_twist,tip_section_twist]) 

        # if wing is a vertical wing, the y and z coordinates are swapped 
        if avl_wing.vertical:
            inverted_wing = -np.sign(abs(dihedral) - np.pi/2)
            if inverted_wing  == 0: inverted_wing  = 1
            dz = ordered_section_spans[section_count] -  inverted_wing*semispan*root_percent_span
            dy = dz*np.tan(dihedral)
            l  = dz/np.cos(dihedral)
            dx = l*np.tan(sweep)                                                            
        else:
            inverted_wing = np.sign(dihedral)
            if inverted_wing  == 0: inverted_wing  = 1
            dy = ordered_section_spans[section_count] - inverted_wing*semispan*root_percent_span
            dz = dy*np.tan(dihedral)
            l  = dy/np.cos(dihedral)
            dx = l*np.tan(sweep)
        section.origin = [[origin[0][0] + dx , origin[0][1] + dy, origin[0][2] + dz]]              

        # this loop appends all the control surfaces within a particular wing section
        for index  , ctrl_surf in enumerate(rcaide_wing.control_surfaces):
            if  (semispan*ctrl_surf.span_fraction_start == ordered_section_spans[section_count]) or (ordered_section_spans[section_count] == semispan*ctrl_surf.span_fraction_end):
                c                     = Control_Surface()
                c.tag                 = ctrl_surf.tag                # name of control surface   
                c.sign_duplicate      = '+1'                         # this float indicates control surface deflection symmetry
                c.x_hinge             = 1 - ctrl_surf.chord_fraction # this float is the % location of the control surface hinge on the wing 
                c.deflection          = ctrl_surf.deflection / Units.degrees 
                c.order               = index

                # if control surface is an aileron, the deflection is asymmetric. This is standard convention from AVL
                if (type(ctrl_surf) ==  Aileron):
                    c.sign_duplicate = '-1'
                    c.function       = 'aileron'
                    c.gain           = -1.0
                # if control surface is a slat, the hinge is taken from the leading edge        
                elif (type(ctrl_surf) ==  Slat):
                    c.x_hinge   =  -ctrl_surf.chord_fraction
                    c.function  = 'slat'
                    c.gain      = -1.0
                elif (type(ctrl_surf) ==  Flap):
                    c.function  = 'flap'    
                    c.gain      = 1.0
                elif (type(ctrl_surf) ==  Elevator):
                    c.function  = 'elevator'
                    c.gain      = 1.0
                elif (type(ctrl_surf) ==  Rudder):
                    c.function  = 'rudder'
                    c.gain      = 1.0
                else:
                    raise AttributeError("Define control surface function as 'slat', 'flap', 'elevator' , 'aileron' or 'rudder'")
                section.append_control_surface(c) 

            elif  (semispan*ctrl_surf.span_fraction_start < ordered_section_spans[section_count]) and (ordered_section_spans[section_count] < semispan*ctrl_surf.span_fraction_end):
                c                     = Control_Surface()
                c.tag                 = ctrl_surf.tag                # name of control surface   
                c.sign_duplicate      = '+1'                         # this float indicates control surface deflection symmetry
                c.x_hinge             = 1 - ctrl_surf.chord_fraction # this float is the % location of the control surface hinge on the wing 
                c.deflection          = ctrl_surf.deflection / Units.degrees 
                c.order               = index

                # if control surface is an aileron, the deflection is asymmetric. This is standard convention from AVL
                if (type(ctrl_surf) ==  Aileron):
                    c.sign_duplicate = '-1'
                    c.function       = 'aileron'
                    c.gain           = -1.0
                # if control surface is a slat, the hinge is taken from the leading edge        
                elif (type(ctrl_surf) ==  Slat):
                    c.x_hinge   =  -ctrl_surf.chord_fraction
                    c.function  = 'slat'
                    c.gain      = -1.0
                elif (type(ctrl_surf) ==  Flap):
                    c.function  = 'flap'    
                    c.gain      = 1.0
                elif (type(ctrl_surf) ==  Elevator):
                    c.function  = 'elevator'
                    c.gain      = 1.0
                elif (type(ctrl_surf) ==  Rudder):
                    c.function  = 'rudder'
                    c.gain      = 1.0
                else:
                    raise AttributeError("Define control surface function as 'slat', 'flap', 'elevator' , 'aileron' or 'rudder'")
                section.append_control_surface(c)                                                  

        if isinstance(tip_airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):  
            if type(tip_airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil:
                section.naca_airfoil         = tip_airfoil.NACA_4_Series_code 
            else:
                section.airfoil_coord_file   = write_avl_airfoil_file(tip_airfoil.coordinate_file) 

        avl_wing.append_section(section)  
                        
    return 
def populate_body_sections(avl_body,rcaide_body):
    """ Creates sections of body geometry and populates the AVL body data structure

    Assumptions:
        None

    Source:
        None

    Inputs:
        avl_wing.symmetric                       [boolean]
        avl_body.widths.maximum                  [meters]
        avl_body.heights.maximum                 [meters]
        rcaide_body.fineness.nose                 [meters]
        rcaide_body.fineness.tail                 [meters]
        avl_body.lengths.total                   [meters]
        avl_body.lengths.nose                    [meters] 
        avl_body.lengths.tail                    [meters]  

    Outputs:
        avl_body - aircraft body in AVL format   [data stucture] 

    Properties Used:
        N/A
    """  

    symm = avl_body.symmetric   
    semispan_h = avl_body.widths.maximum * 0.5 * (2 - symm)
    semispan_v = avl_body.heights.maximum * 0.5
    origin = rcaide_body.origin[0]

    # Compute the curvature of the nose/tail given fineness ratio. Curvature is derived from general quadratic equation
    # This method relates the fineness ratio to the quadratic curve formula via a spline fit interpolation
    vec1 = [2 , 1.5, 1.2 , 1]
    vec2 = [1  ,1.57 , 3.2,  8]
    x = np.linspace(0,1,4)
    fuselage_nose_curvature =  np.interp(np.interp(rcaide_body.fineness.nose,vec2,x), x , vec1)
    fuselage_tail_curvature =  np.interp(np.interp(rcaide_body.fineness.tail,vec2,x), x , vec1) 


    # Horizontal Sections of Fuselage
    if semispan_h != 0.0:                
        width_array = np.linspace(-semispan_h, semispan_h, num=11,endpoint=True)
        for section_width in width_array:
            fuselage_h_section               = Section()
            fuselage_h_section_cabin_length  = avl_body.lengths.total - (avl_body.lengths.nose + avl_body.lengths.tail)
            fuselage_h_section_nose_length   = ((1 - ((abs(section_width/semispan_h))**fuselage_nose_curvature ))**(1/fuselage_nose_curvature))*avl_body.lengths.nose
            fuselage_h_section_tail_length   = ((1 - ((abs(section_width/semispan_h))**fuselage_tail_curvature ))**(1/fuselage_tail_curvature))*avl_body.lengths.tail
            fuselage_h_section_nose_origin   = avl_body.lengths.nose - fuselage_h_section_nose_length
            fuselage_h_section.tag           =  'fuselage_horizontal_section_at_' +  str(section_width) + '_m'
            fuselage_h_section.origin        = [ origin[0] + fuselage_h_section_nose_origin , origin[1] + section_width, origin[2]]
            fuselage_h_section.chord         = fuselage_h_section_cabin_length + fuselage_h_section_nose_length + fuselage_h_section_tail_length
            avl_body.append_section(fuselage_h_section,'horizontal')

    # Vertical Sections of Fuselage 
    if semispan_v != 0:               
        height_array = np.linspace(-semispan_v, semispan_v, num=11,endpoint=True)
        for section_height in height_array :
            fuselage_v_section               = Section()
            fuselage_v_section_cabin_length  = avl_body.lengths.total - (avl_body.lengths.nose + avl_body.lengths.tail)
            fuselage_v_section_nose_length   = ((1 - ((abs(section_height/semispan_v))**fuselage_nose_curvature ))**(1/fuselage_nose_curvature))*avl_body.lengths.nose
            fuselage_v_section_tail_length   = ((1 - ((abs(section_height/semispan_v))**fuselage_tail_curvature ))**(1/fuselage_tail_curvature))*avl_body.lengths.tail
            fuselage_v_section_nose_origin   = avl_body.lengths.nose - fuselage_v_section_nose_length
            fuselage_v_section.tag           = 'fuselage_vertical_top_section_at_' +  str(section_height) + '_m'        
            fuselage_v_section.origin        = [ origin[0] + fuselage_v_section_nose_origin,  origin[1],  origin[2] + section_height ]
            fuselage_v_section.chord         = fuselage_v_section_cabin_length + fuselage_v_section_nose_length + fuselage_v_section_tail_length
            avl_body.append_section(fuselage_v_section,'vertical')

    return avl_body




# RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/evaluate_cantera.py

# Created: June 2024, M. Clarke, M. Guidotti 
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from   RCAIDE.Framework.Core import Data  
import pandas                as pd
import numpy                 as np
import os

# ----------------------------------------------------------------------------------------------------------------------
#  evaluate_cantera
# ----------------------------------------------------------------------------------------------------------------------   
def evaluate_cantera(combustor,T,P,mdot,FAR): 

    """
    Evaluates emission indices using Chemical Reactor Network (CRN) built in Cantera.
    
    Parameters
    ----------
    combustor : Data
        Combustor configuration data
        
        - f_air_PZ : float
            Fraction of total air entering Primary Zone [-]
        - FAR_st : float
            Stoichiometric Fuel to Air ratio [-]
        - N_comb : int
            Number of can-annular combustors [-]
        - N_PZ : int
            Number of PSR (Perfectly Stirred Reactors) [-]
        - A_PZ : float
            Primary Zone cross-sectional area [m^2]
        - L_PZ : float
            Primary Zone length [m]
        - N_SZ : int
            Number of dilution air inlets [-]
        - A_SZ : float
            Secondary Zone cross-sectional area [m^2]
        - L_SZ : float
            Secondary Zone length [m]
        - phi_SZ : float
            Equivalence Ratio in the Secondary Zone [-]
        - S_PZ : float
            Mixing parameter for Primary Zone [-]
        - F_SC : float
            Fuel scaler [-]
        - number_of_assigned_PSR_1st_mixers : int
            Number of PSR assigned to first row of mixers [-]
        - number_of_assigned_PSR_2nd_mixers : int
            Number of PSR assigned to second row of mixers [-]
        - fuel_data : Data
            Fuel chemical properties and kinetics data

    T : float
        Stagnation Temperature entering combustors [K]
    P : float
        Stagnation Pressure entering combustors [Pa]
    mdot : float
        Air mass flow enetring the combustor [kg/s]
    FAR : float
        Fuel-to-Air ratio [-]

    Returns
    -------
    results : Data
        Container for emission indices
        
        - EI_CO2 : float
            CO2 emission index [kg_CO2/kg_fuel]
        - EI_CO : float
            CO emission index [kg_CO/kg_fuel]
        - EI_H2O : float
            H2O emission index [kg_H2O/kg_fuel]
        - EI_NO : float
            NO emission index [kg_NO/kg_fuel]
        - EI_NO2 : float
            NO2 emission index [kg_NO2/kg_fuel]

    Notes
    -----
    This model estimates emission indices using a Chemical Reactor Network built with Cantera.
    The network consists of Perfectly Stirred Reactors (PSRs) in the Primary Zone and 
    Plug Flow Reactors (PFRs) in the Secondary Zone.

    **Extra modules required**

    * Cantera
    * pandas
    * numpy

    **Major Assumptions**

    * PSRs represent the Primary Zone
    * Air is added between different PFRs in Secondary Zone
    * Primary Zone has a normal Equivalence Ratio distribution

    **Theory**
    The model uses a network of reactors to simulate combustion:
    
    1. Primary Zone: Multiple PSRs with varying equivalence ratios
    2. Mixing Zones: Ideal mixing between reactor outputs
    3. Secondary Zone: PFRs with dilution air addition

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_CRN_emission_indices
    RCAIDE.Library.Components.Propulsors.Converters.Combustor

    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics, thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    """   

    # ------------------------------------------------------------------------------              
    # ------------------------------ Combustor Inputs ------------------------------              
    # ------------------------------------------------------------------------------              
    
    high_fidelity_kin_mech  = False                                                 # [-]       True (simulation time around 300 s): Computes EI_CO2, EI_CO, EI_H2O, EI_NO2, EI_NO, EI_CSOOT; False (simulation time around 60 s): Computes EI_CO2, EI_CO, EI_H2O
    T_stag_0                = T                                                     # [K]       Stagnation Temperature entering all combustors
    P_stag_0                = P                                                     # [Pa]      Stagnation Pressure entering all combustors
    FAR                     = FAR                                                   # [-]       Fuel-to-Air ratio
    m_dot_air_tot           = mdot                                                  # [kg/s]    Air mass flow going through all combustors
    
    f_air_PZ                = combustor.f_air_PZ                                    # [-]       Fraction of total air present in the combustor that enters the Primary Zone         
    FAR_st                  = combustor.FAR_st                                      # [-]       Stoichiometric Fuel to Air ratio 
    N_comb                  = combustor.N_comb                                      # [-]       Number of can-annular combustors
    N_PZ                    = combustor.N_PZ                                        # [-]       Number of PSR (EVEN, must match the number of PSR below)
    A_PZ                    = combustor.A_PZ                                        # [m**2]    Primary Zone cross-sectional area     
    L_PZ                    = combustor.L_PZ                                        # [m]       Primary Zone length  
    N_SZ                    = combustor.N_SZ                                        # [-]       Number of dilution air inlets        
    A_SZ                    = combustor.A_SZ                                        # [m**2]    Secondary Zone cross-sectional area
    L_SZ                    = combustor.L_SZ                                        # [m]       Secondary Zone length  
    phi_SZ                  = combustor.phi_SZ                                      # [-]       Equivalence Ratio for PFR    phi_PZ_des              = 0.6                                                   # [-]       Primary Zone Design Equivalence Ratio
    S_PZ                    = combustor.S_PZ                                        # [-]       Mixing parameter, used to define the Equivalence Ratio standard deviation  
    F_SC                    = combustor.F_SC                                        # [-]       Fuel scaler
    PSR_1st_mixers          = combustor.number_of_assigned_PSR_1st_mixers           # [-]       Number of PSR assigned to the first row of mixers  
    PSR_2nd_mixers          = combustor.number_of_assigned_PSR_2nd_mixers           # [-]       Number of mixers assigned to the second row of mixers                                           
         
    m_dot_fuel_tot          = m_dot_air_tot*FAR                                     # [kg/s]    Fuel mass flow going through all combustors
    m_dot_air               = m_dot_air_tot/N_comb                                  # [kg/s]    Air mass flow inside each combustor, scaled inside each PSR to vary the Equivalence Ratio
    m_dot_fuel              = m_dot_fuel_tot/N_comb                                 # [kg/s]    Fuel mass flow inside each combustor                                                                                                                                                                    
    phi_sign                = (m_dot_fuel_tot*F_SC)/(m_dot_air_tot*f_air_PZ*FAR_st) # [-]       Primary Zone mean Equivalence Ratio
    sigma_phi               = phi_sign*S_PZ                                         # [-]       Primary Zone Equivalence Ratio standard deviation                                                                                                                                                      
    
    if high_fidelity_kin_mech:                                                               
        dict_fuel           = combustor.fuel_data.fuel_chemical_properties          # [-]       Fuel species and corresponding mole fractions for full fuel model
        kinetics_model      = combustor.fuel_data.surrogate_chemical_kinetics       # [-]       Full fuel model 
    else:                                                                                      
        dict_fuel           = combustor.fuel_data.fuel_surrogate_chemical_properties# [-]       Fuel species and corresponding mole fractions for surrogate fuel model 
        kinetics_model      = combustor.fuel_data.surrogate_chemical_kinetics       # [-]       Surrogate fuel model 
    dict_oxy                = combustor.fuel_data.air_chemical_properties           # [-]       Air species and corresponding mole fractions     
    oxidizer_model          = combustor.fuel_data.oxidizer
    
    if high_fidelity_kin_mech:                                                                           
        list_sp             = combustor.fuel_data.species_list                      # [-]       Fuel species for Emission Index analysis on the full model
    else:                                                                                       
        list_sp             = combustor.fuel_data.surrogate_species_list            # [-]       Fuel species for Emission Index analysis on the surrogate model
    
    col_names               = ['EI_' +str(sp) for sp in list_sp]                    # [-]       Define output variables 
    df                      = pd.DataFrame(columns=col_names)                       # [-]       Assign output variables space to df
    
    
    try: 
        import cantera as ct 
        gas, EI = combustor_model(ct,kinetics_model,oxidizer_model,high_fidelity_kin_mech, dict_fuel, dict_oxy, T_stag_0, P_stag_0, FAR, m_dot_fuel, m_dot_air, N_PZ, A_PZ, L_PZ, phi_sign, phi_SZ, A_SZ, L_SZ, f_air_PZ, N_SZ, sigma_phi, PSR_1st_mixers, PSR_2nd_mixers) # [-]       Run combustor function
        
        sp_idx                  = [gas.species_index(sp) for sp in list_sp]             # [-]       Retrieve the species index
        data_n                  = list(EI[sp_idx])                                      # [-]       Assign output variables  
        df.loc[0]               = data_n                                                # [-]       Assign output variables to df 
        
        results                 = Data()                                                # [-]       Create results data structure
        results.EI_CO2          = df.loc[0, 'EI_CO2']                                   # [-]       Assign CO2 Emission Index to the results data structure
        results.EI_CO           = df.loc[0, 'EI_CO']                                    # [-]       Assign CO Emission Index to the results data structure
        results.EI_H2O          = df.loc[0, 'EI_H2O']                                   # [-]       Assign H2O Emission Index to the results data structure
        results.EI_NO           = 0                                                     # [-]       Assign NO Emission Index to the results data structure
        results.EI_NO2          = 0                                                     # [-]       Assign NO2 Emission Index to the results data structure   
        
        if high_fidelity_kin_mech:   
            results.EI_NO       = df.loc[0, 'EI_NO']                                    # [-]       Assign NO Emission Index to the results data structure
            results.EI_NO2      = df.loc[0, 'EI_NO2']                                   # [-]       Assign NO2 Emission Index to the results data structure
 
    except ImportError:
        print('cantera required: run pip install cantera') 
        results                 = Data()                                               
        results.EI_CO2          = 0                               
        results.EI_CO           = 0                               
        results.EI_H2O          = 0                               
        results.EI_NO           = 0                                                    
        results.EI_NO2          = 0                                                    
        
        if high_fidelity_kin_mech:   
            results.EI_NO       = 0                                 
            results.EI_NO2      = 0                                 
     

    return results

def combustor_model(ct, kinetics_model,oxidizer_model,high_fidelity_kin_mech, dict_fuel, dict_oxy, T_stag_0, P_stag_0, FAR, m_dot_fuel, m_dot_air, N_PZ, A_PZ, L_PZ, phi_sign, phi_SZ, A_SZ, L_SZ, f_air_PZ, N_SZ, sigma_phi, PSR_1st_mixers, PSR_2nd_mixers):

    ospath    = os.path.abspath(__file__)
    separator = os.path.sep
    rel_path  = os.path.dirname(ospath) + separator + "Data" + separator
    

    # ------------------------------------------------------------------------------
    # ----------------------------- Initial Parameters -----------------------------
    # ------------------------------------------------------------------------------    
      
    f_air_SZ                = 1 - f_air_PZ                                          # [-]       Fraction of total air present in the combustor that enters the Secondary Zone  
    m_dot_air_PSR           = f_air_PZ*m_dot_air                                    # [kg/s]    Air mass flow going through each PSR  
    m_dot_fuel_PSR          = m_dot_fuel                                            # [kg/s]    Fuel mass flow going through each PSR    
    m_dot_air_SZ            = (f_air_SZ*m_dot_air)/N_SZ                             # [kg/s]    Air mass flow going through each dilution air inlet (3 inlets)
    V_PZ_PSR                = (A_PZ*L_PZ)/N_PZ                                      # [m**3]    Volume of each PSR
    phi_PSR                 = np.linspace(0.001, 2*phi_sign, N_PZ)                  # [-]       Distribution of Equivalence Ratio through the PSRs
    Delta_phi               = np.abs(phi_PSR[0] - phi_PSR[1])                       # [-]       Difference between two subsequent Equivalence Ratios
   
    # ------------------------------------------------------------------------------
    # ----------------------------------- PSR #1 -----------------------------------
    # ------------------------------------------------------------------------------ 

    Fuel_list               = []                                                    # [-]       Create a list for the fuel inside each PSR
    mass_flow_rates         = []                                                    # [-]       Create a list for the mass flow rate inside each PSR
    mixer_list              = []                                                    # [-]       Create a list for the mixers inside each PSR
    
    m_i = 0                                                                         # [-]       Initialize the index of the mixer
       
    for PSR_i in range(N_PZ):     
        f_PZ_1                            = (1 / (np.sqrt(2 * np.pi) * sigma_phi)) * np.exp((-(phi_PSR[PSR_i] - phi_sign) ** 2) / (2 * sigma_phi ** 2)) * Delta_phi # [-]       Fraction of mass flow entering the PSR at the PSR equivalence ratio                                                                                       
        Fuel_list.append(ct.Solution(rel_path+kinetics_model) )                     # [-]       Import surrogate fuel kinematic mechanism
        Fuel_list[PSR_i].TP               = T_stag_0, P_stag_0                      # [-]       Set the fuel temperature and pressure
        Fuel_list[PSR_i].set_equivalence_ratio(phi_PSR[PSR_i], fuel=dict_fuel, oxidizer=dict_oxy) # [-]       Set the euivalence ratio inside the PSR
        Fuel_list[PSR_i].equilibrate('HP')                                          # [-]       Fix the specific enthalpy and pressure 
        rho                               = Fuel_list[PSR_i].density                # [kg/m**3] Fuel density
        m_dot_air_PSRl                         = m_dot_air_PSR * f_PZ_1             # [kg/s]    Air mass flow inside the PSR    
        m_dot_fuel_PSRl                        = m_dot_fuel_PSR * f_PZ_1            # [kg/s]    Fuel mass flow inside the PSR
        mass_flow_rates.append(m_dot_fuel_PSRl + m_dot_air_PSRl)                    # [kg/s]    Total mass flow inside the PSR  
        upstream                          = ct.Reservoir(Fuel_list[PSR_i])          # [-]       Create a resevoir for the Fuel upstream of the PSR
        if PSR_i % PSR_1st_mixers == 0:
            mixer_list.append(ct.IdealGasReactor(Fuel_list[PSR_i]))                 # [-]       Create the reactor for the downstream mixer
        
        PSR                               = ct.IdealGasReactor(Fuel_list[PSR_i])    # [-]       Create the reactor for the PSR 
        PSR.volume                        = V_PZ_PSR                                # [m**3]    Set the PSR volume
        inlet                             = ct.MassFlowController(upstream, PSR)    # [-]       Connect the upstream resevoir with the PSR 
        inlet.mass_flow_rate              = mass_flow_rates[PSR_i]                  # [kg/s]    Prescribe the inlet mass flow rate 
        outlet                            = ct.MassFlowController(PSR, mixer_list[m_i], mdot=mass_flow_rates[PSR_i]) # [-]       Connect the PSR with the downstream mixer with the same mass flow rate                                     
        t_res_PSR                         = (rho * V_PZ_PSR) / (mass_flow_rates[PSR_i]) # [s]       Compute the residence time inside the PSR
        sim_PSR                           = ct.ReactorNet([PSR])                    # [-]       Set the PSR simulation
        sim_PSR.advance(t_res_PSR)                                                  # [-]       Run the simulation until the residence time is reached 
           
        if PSR_i % PSR_1st_mixers != 0:                                             # [-]       If the number of PSR / number of PSRs assigned to the mixer is different from 0
            m_i += 1                                                                # [-]       Increase the index of the mixer
        
    mass_flow_rates =  np.array(mass_flow_rates)                                    # [-]       Create array of mass flow rates
    mixer_list_2    =  []                                                           # [-]       Initialize the list of mixers in the second talk
    idx   =  0                                                                      # [-]       Initialize the first index
    idx_2 =  0                                                                      # [-]       Initialize the second index
    for mixer_i in range(0, N_PZ, int(N_PZ/PSR_2nd_mixers)): 
        mixer_list_2.append(ct.IdealGasReactor(Fuel_list[mixer_i]))                 # [-]       Create the reactor for the downstream mixer        
        for n in range(int(N_PZ/PSR_1st_mixers/PSR_2nd_mixers)):
            start_index =  mixer_i +  n * 2                                         # [-]       Define the starting index
            end_index   =  mixer_i + ( n + 1 ) * 2                                  # [-]       Define the ending index
            outlet      =  ct.MassFlowController(mixer_list[idx_2], mixer_list_2[idx], mdot = np.sum(mass_flow_rates[start_index:end_index])) # [-]       Connect the mixer with the downstream mixer     
            sim_mixer   =  ct.ReactorNet([mixer_list[idx_2]])                       # [-]       Set the mixer simulation  
            sim_mixer.advance_to_steady_state()                                     # [-]       Run the simulation until it reaches steady state
            idx_2 += 1                                                              # [-]       Advance second index of one
        idx +=  1                                                                   # [-]       Advance first index of one
     
    # ------------------------------------------------------------------------------
    # ------------------------------- Mixing 1-2-3-4 -------------------------------
    # ------------------------------------------------------------------------------
    
    mixer_12345678  = ct.IdealGasReactor(Fuel_list[0])                              # [-]       Create the reactor for the downstream mixer
    for idx_3 in  range(2):
        start_index = idx_3 * 4                                                     # [-]       Define the starting index
        end_index = (idx_3 + 1) * 4                                                 # [-]       Define the ending index   
        outlet_1234             = ct.MassFlowController(mixer_list_2[idx_3], mixer_12345678, mdot = np.sum(mass_flow_rates[ start_index:end_index]) ) # [-]       Connect the mixer with the downstream mixer     
        sim_mixer_1234          = ct.ReactorNet([mixer_list_2[idx_3]])              # [-]       Set the mixer simulation  
        sim_mixer_1234.advance_to_steady_state()                                    # [-]       Run the simulation until it reaches steady state
         
    # ------------------------------------------------------------------------------
    # --------------------------- Mixing 1-2-3-4-5-6-7-8 ---------------------------
    # ------------------------------------------------------------------------------     
  
    mixer_air         = ct.IdealGasReactor(Fuel_list[0])                            # [-]       Create the reactor for the downstream mixer
    outlet_12345678   = ct.MassFlowController(mixer_12345678, mixer_air, mdot = (  np.sum(np.array(mass_flow_rates)))) # [-]       Connect the mixer with the downstream mixer     
    sim_mixer_12345678      = ct.ReactorNet([mixer_12345678])                       # [-]       Set the mixer simulation  
    sim_mixer_12345678.advance_to_steady_state()                                    # [-]       Run the simulation until it reaches steady state

    # ------------------------------------------------------------------------------
    # -------------------------------- Mixing 1-Air --------------------------------
    # ------------------------------------------------------------------------------
    
    total_mass_flow = np.sum(np.array(mass_flow_rates))                             # [kg/s]    Total mass flow entering the first mixer  

    for _ in  range(N_SZ): 
        Air                   = ct.Solution(rel_path+oxidizer_model)                # [-]       Import air kinematic mechanism           
        Air.TPX               = T_stag_0, P_stag_0, dict_oxy                        # [-]       Set the air temperature, pressure and mole fractions
        rho_air               = Air.density                                         # [kg/m**3] Fuel density
        res_air               = ct.Reservoir(Air)                                   # [-]       Create a resevoir for the air upstream of the mixer
        PFR                   = ct.IdealGasConstPressureReactor(Fuel_list[0])       # [-]       Create the reactor for the PFR
        PFR.volume            = A_SZ*(L_SZ/N_SZ)                                    # [m**3]    Set the PFR volume
        inlet_air             = ct.MassFlowController(res_air, mixer_air, mdot= m_dot_air_SZ) # [-]       Connect the upstream resevoir with the mixer
        total_mass_flow       = total_mass_flow + m_dot_air_SZ                      # [kg/s]    Total mass flow inside the mixer  
        outlet_air            = ct.MassFlowController(mixer_air, PFR, mdot=total_mass_flow ) # [-]       Connect the mixer with the downstream PFR with the same mass flow rate           
        sim_mixer_air         = ct.ReactorNet([mixer_air])                          # [-]       Set the mixer simulation  
        sim_mixer_air.advance_to_steady_state()                                     # [-]       Run the simulation until it reaches steady state
        
        # --------------------------------------------------------------------------
        # ------------------------------------ PFR ---------------------------------
        # -------------------------------------------------------------------------- 
        
        mixer_air             = ct.IdealGasReactor(Fuel_list[0])                    # [-]       Create the reactor for the downstream mixer
        outlet_PFR            = ct.MassFlowController(PFR,mixer_air , mdot=total_mass_flow ) # [-]       Connect the PFR with the downstream mixer                                 
        t_res_PFR             = (Fuel_list[0].density * PFR.volume) / (total_mass_flow) # [s]       Compute the residence time inside the PFR           
        Fuel_list[0].set_equivalence_ratio(phi_SZ, fuel=dict_fuel, oxidizer=dict_oxy) # [-]       Set the euivalence ratio inside the PFR
        sim_PFR               = ct.ReactorNet([PFR])                                # [-]       Set the PFR simulation
        sim_PFR.advance(t_res_PFR)                                                  # [-]       Run the simulation until the residence time is reached 
   
    # ------------------------------------------------------------------------------
    # ---------------------------- Additional computations -------------------------
    # ------------------------------------------------------------------------------    

    m_dot_input_combustor   = m_dot_fuel + m_dot_air                                # [kg/s]    Total mass flow rate entering a single combustor (air + fuel)
    Emission_Index          = Fuel_list[0].Y * (m_dot_input_combustor)/m_dot_fuel   # [-]       Computation of the Emission Index   

    return (Fuel_list[0], Emission_Index)  


# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/write_geometry.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Units
from .purge_files import purge_files

# ---------------------------------------------------------------------------------------------------------------------- 
# Write Input Deck
# ----------------------------------------------------------------------------------------------------------------------  
def write_input_deck(dfdc_object):
    """ This function writes the execution steps used in the DFDC call 
    """
    # unpack 
    deck_filename = dfdc_object.current_status.deck_file 

    # purge old versions and write the new input deck
    purge_files([deck_filename]) 
    with open(deck_filename,'w') as input_deck:
        
        header_text     = make_header_text(dfdc_object)
        input_deck.write(header_text)
        
        settings_text     = make_settings_text(dfdc_object)
        input_deck.write(settings_text)
        
        for case in dfdc_object.run_cases:
            # write and store aerodynamic and static stability result files 
            case_command = make_case_command(dfdc_object,case)
            input_deck.write(case_command) 
        input_deck.write('\nQUIT\n')

    return

def make_header_text(dfdc_object): 
    base_header_text= \
'''LOAD
{0}
OPER
'''  
    header_text =  base_header_text.format(dfdc_object.settings.filenames.case)
    return header_text


def make_settings_text(dfdc_object):
    """ Makes commands for case execution in DFDC
    """  
    # This is a template (place holder) for the input deck. Think of it as the actually keys
    # you will type if you were to manually run an analysis
    base_settings_command = \
'''atmo
{0}
vref
{1}
vinf
{2}
nbld 
{3}
nrs 
{4}
rpm
{5}
thru
{6}
desi
shoo
{7}
'''
    ducted_fan        = dfdc_object.geometry 
    geometry_filename = ducted_fan.tag + '_geometry.txt'                    
    B                 = ducted_fan.number_of_rotor_blades         
    n                 = ducted_fan.number_of_radial_stations + 1           
    T                 = ducted_fan.cruise.design_thrust               
    alt               = ducted_fan.cruise.design_altitude /1000     
    RPM               = ducted_fan.cruise.design_angular_velocity /Units.rpm   
    V_inf             = ducted_fan.cruise.design_freestream_velocity  
    V_ref             = ducted_fan.cruise.design_reference_velocity  
    settings_command  = base_settings_command.format(alt,V_ref,V_inf,B,n,RPM,T,geometry_filename) 
 
    if not dfdc_object.settings.keep_files:
        purge_files([geometry_filename])         
    return settings_command

def make_case_command(dfdc_object,case):
    """ Makes commands for case execution in DFDC
    """  
    # This is a template (place holder) for the input deck. Think of it as the actually keys
    # you will type if you were to manually run an analysis
    base_case_command = \
'''atmo
{0}
rpm
{1} 
vinf
{2}
exec
writ
N
{3}
'''    
    alt               = case.altitude   
    V_inf             = case.velocity     
    RPM               = case.RPM        
    results_filename  = case.tag
    case_command      = base_case_command.format(alt,RPM,V_inf,results_filename)  
        
    return case_command 